# import ray
# import numpy as np
# import pandas as pd
# from ncls import NCLS

# import pyranges as pr

# from natsort import natsorted

# from joblib import Parallel, delayed

# from sorted_nearest import (find_clusters, nearest_previous_nonoverlapping,
#                             nearest_next_nonoverlapping, nearest_nonoverlapping, find_clusters)

# from collections import defaultdict

# from functools import wraps

# def parse_grpby_key(grpby_key):

#     if isinstance(grpby_key, str):
#         return grpby_key, False
#     else:
#         return grpby_key[0], grpby_key[1]


# def return_empty_if_one_empty(func):

#     @wraps(func)
#     def extended_func(self, other, **kwargs):

#         if len(self) == 0 or len(other) == 0:
#             df = pd.DataFrame(columns="Chromosome Start End".split())
#         else:
#             df = func(self, other, **kwargs)

#         return df

#     return extended_func


# def _pyrange_apply(function, self_dfs, other_dfs, grpby_key, **kwargs):

#     # print("in pyrange apply " * 10, n_jobs)
#     # if function.__name__ == "_set_union":
#     #     self_dfs =  {k: d for k, d in scdf.groupby(grpby_key)}
#     #     self_dfs = defaultdict(lambda: pd.DataFrame(columns="Chromosome Start End".split()), self_dfs)
#     #     keys_union = natsorted(list(set(self_dfs).union(other_dfs)))
#     #     outdfs = Parallel(n_jobs=n_jobs)(delayed(function)(self_dfs[key], other_dfs[key], key=key, n_jobs=n_jobs, **kwargs) for key in keys_union)

#     # elif not function.__name__ in ["_write_both", "_nearest"]:
#     #     # for most methods, we do not need to know anything about other except start, end
#     #     # i.e. less data that needs to be sent to other processes
#     #     outdfs = Parallel(n_jobs=n_jobs)(delayed(function)(scdf, other_dfs[key][["Start", "End"]], key=key, n_jobs=n_jobs, **kwargs) for key, scdf in natsorted(scdf.groupby(grpby_key)))

#     # else:
#         # outdfs = Parallel(n_jobs=n_jobs)(delayed(function)(scdf, other_dfs[key], key=key, n_jobs=n_jobs, **kwargs) for key, scdf in natsorted(scdf.groupby(grpby_key)))
#     results = []
#     for key, scdf in natsorted(self_dfs.items()):
#         # print("result: ", _intersection(scdf, other_dfs[key]))
#         # result = function.remote()
#         # result = function.remote(scdf=scdf, ocdf=other_dfs[key])
#         # result = pr.ray._intersection.remote()
#         result = pr.ray._intersection.remote(scdf=scdf, ocdf=other_dfs[key], how=kwargs["how"])

#         results.append(result)

#     outdfs = ray.get(results)

#     return {k: v  for k, v in zip(natsorted(self_dfs), outdfs)}

#     # PyRanges(ou)


# def pyrange_apply_single(function, self, **kwargs):

#     strand = kwargs["strand"]
#     n_jobs = kwargs.get("n_jobs", 1)

#     print("Using {} cores and strand {}".format(n_jobs, strand))

#     if strand:
#         assert self.stranded, \
#             "Can only do stranded operation when PyRange contains strand info"

#     if self.stranded and strand:
#         grpby_key = ["Chromosome", "Strand"]
#     else:
#         grpby_key = "Chromosome"

#     outdfs = Parallel(n_jobs=n_jobs)(delayed(function)(scdf, **kwargs) for key, scdf in natsorted(self.df.groupby(grpby_key)))


#     outdfs = [df for df in outdfs if not df.empty]

#     if outdfs:
#         df = pd.concat(outdfs)
#         return df
#     else:
#         return pd.DataFrame(columns="Chromosome Start End Strand".split())


# def pyrange_apply(function, self, other, **kwargs):

#     strandedness = kwargs["strandedness"]

#     print("Strandedness {}".format(strandedness))
#     # print("Function name {}".format(function.name))


#     assert strandedness in ["same", "opposite", False, None]

#     if strandedness:
#         assert self.stranded and other.stranded, \
#             "Can only do stranded operations when both PyRanges contain strand info"

#     if self.stranded and other.stranded and strandedness:
#         grpby_key = ["Chromosome", "Strand"]
#     else:
#         grpby_key = "Chromosome"

#     other_strand = {"+": "-", "-": "+"}
#     if strandedness == "opposite":
#         other_dfs = {(c, other_strand[s]): v for (c, s), v in other.dfs.items()}
#         other_dfs = defaultdict(lambda: pd.DataFrame(columns="Chromosome Start End Strand".split()), other_dfs)
#         # other_ncls = {(c, other_strand[s]): v for (c, s), v in other.__ncls__.items()}
#         # other_ncls = defaultdict(NCLS)
#     else:
#         other_dfs = {(c, other_strand[s]): v for (c, s), v in other.dfs.items()}
#         other_dfs = defaultdict(lambda: pd.DataFrame(columns="Chromosome Start End".split()), other_dfs)
#         # other_ncls = {(c, other_strand[s]): v for (c, s), v in other.__ncls__.items()}
#         # other_ncls = defaultdict(NCLS, other_ncls)

#     return _pyrange_apply(function, self.dfs, other_dfs, grpby_key, **kwargs)




# @return_empty_if_one_empty
# def _first_df(scdf, ocdf, how=False, invert=False, n_jobs=1, **kwargs):

#     assert how in "containment first".split() + [False, None]
#     starts = scdf.Start.values
#     ends = scdf.End.values
#     indexes = scdf.index.values

#     print("n_jobs " * 10)
#     print(n_jobs)

#     if n_jobs > 1:
#         print("deepcopy")
#         scdf = scdf.copy(deep=True)

#     it = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

#     if not how:
#         _indexes = it.has_overlaps(starts, ends, indexes)
#     elif how == "containment":
#         _indexes = it.has_containments(starts, ends, indexes)

#     if not invert:
#         return scdf.reindex(_indexes)
#     else:
#         return scdf.loc[~scdf.index.isin(_indexes)]


# # @ray.remote(num_return_vals=2)
# # def _both_dfs(scdf, ocdf, how=False):

# #     assert how in "containment first".split() + [False, None]
# #     starts = scdf.Start.values
# #     ends = scdf.End.values
# #     indexes = scdf.index.values

# #     oncls = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

# #     if not how:
# #         _self_indexes, _other_indexes = oncls.all_overlaps_both(starts, ends, indexes)
# #     elif how == "containment":
# #         _self_indexes, _other_indexes = oncls.all_containments_both(starts, ends, indexes)
# #     else:
# #         _self_indexes, _other_indexes = oncls.first_overlap_both(starts, ends, indexes)

# #     _self_indexes = _self_indexes
# #     _other_indexes = _other_indexes

# #     return scdf.loc[_self_indexes], ocdf.loc[_other_indexes]



# @ray.remote
# def _intersection(scdf, ocdf, how):

#     if ocdf.empty: # just return empty df
#         return ocdf
#     # scdf2, ocdf2 = ray.get(_both_dfs.remote(scdf, ocdf, how=False))
#     assert how in "containment first".split() + [False, None]
#     starts = scdf.Start.values
#     ends = scdf.End.values
#     indexes = scdf.index.values

#     oncls = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

#     if not how:
#         _self_indexes, _other_indexes = oncls.all_overlaps_both(starts, ends, indexes)
#     elif how == "containment":
#         _self_indexes, _other_indexes = oncls.all_containments_both(starts, ends, indexes)
#     else:
#         _self_indexes, _other_indexes = oncls.first_overlap_both(starts, ends, indexes)

#     _self_indexes = _self_indexes
#     _other_indexes = _other_indexes

#     scdf, ocdf = scdf.loc[_self_indexes], ocdf.loc[_other_indexes]

#     new_starts = pd.Series(
#         np.where(scdf.Start.values > ocdf.Start.values, scdf.Start, ocdf.Start),
#         index=scdf.index, dtype=np.long)

#     new_ends = pd.Series(
#         np.where(scdf.End.values < ocdf.End.values, scdf.End, ocdf.End),
#         index=scdf.index, dtype=np.long)

#     pd.options.mode.chained_assignment = None  # default='warn'
#     scdf.loc[:, "Start"] = new_starts
#     scdf.loc[:, "End"] = new_ends
#     pd.options.mode.chained_assignment = 'warn'

#     return scdf


# def _create_df_from_starts_ends(starts, ends, chromosome, strand=None):

#     nidx = pd.Index(range(len(starts)))
#     if strand:
#         cluster_df = pd.DataFrame({"Chromosome": pd.Series(chromosome, dtype="category", index=nidx),
#                                    "Start": starts, "End": ends,
#                                    "Strand": pd.Series(strand, dtype="category", index=nidx)})
#     else:
#         cluster_df = pd.DataFrame({"Chromosome": pd.Series(chromosome, dtype="category", index=nidx), "Start": starts, "End": ends})

#     return cluster_df


# def _cluster(df, chromosome, strand=False, **kwargs):

#     cdf = df.sort_values("Start")
#     starts, ends = find_clusters(cdf.Start.values, cdf.End.values)

#     cluster_df = _create_df_from_starts_ends(starts, ends, chromosome, strand)

#     return cluster_df


# @return_empty_if_one_empty
# def _set_intersection(scdf, ocdf, strandedness=None, how=None, **kwargs):

#     chromosome, strand = parse_grpby_key(kwargs["key"])

#     strand = True if strandedness else False
#     s = _cluster(scdf, chromosome, strand=strand)
#     o = _cluster(ocdf, chromosome, strand=strand)

#     return _intersection(s, o, strandedness=strandedness, how=how, **kwargs)


# def _overlapping_for_nearest(scdf, ocdf, suffix, n_jobs=1, **kwargs):

#     nearest_df = pd.DataFrame(columns="Chromosome Start End Strand".split())

#     scdf2, ocdf2 = _both_dfs(scdf, ocdf, how="first")

#     if not ocdf2.empty:
#         # only copying data because of the eternal source buffer array is read only problem
#         if n_jobs > 1:
#             scdf = scdf.copy(deep=True)
#             original_idx = scdf.index.copy(deep=True)
#         else:
#             original_idx = scdf.index

#         idxs = scdf2.index
#         original_idx = scdf.index.copy(deep=True)
#         missing_idxs = ~original_idx.isin(idxs)
#         missing_overlap = scdf.index[missing_idxs]

#         df_to_find_nearest_in = scdf.reindex(missing_overlap)

#         odf = ocdf.reindex(ocdf2.index)
#         odf.index = idxs
#         sdf = scdf.reindex(idxs)

#         nearest_df = sdf.join(odf, rsuffix=suffix)
#         nearest_df.insert(nearest_df.shape[1], "Distance", 0)
#     else:
#         df_to_find_nearest_in = scdf

#     return nearest_df, df_to_find_nearest_in


# def _next_nonoverlapping(left_ends, right_starts, right_indexes):

#     left_ends = left_ends.sort_values()
#     right_starts = right_starts.sort_values()
#     r_idx, dist = nearest_next_nonoverlapping(left_ends.values - 1, right_starts.values, right_indexes)
#     r_idx = pd.Series(r_idx, index=left_ends.index).sort_index().values
#     dist = pd.Series(dist, index=left_ends.index).sort_index().values

#     return r_idx, dist


# def _previous_nonoverlapping(left_starts, right_ends):

#     left_starts = left_starts.sort_values()
#     right_ends = right_ends.sort_values()
#     r_idx, dist = nearest_previous_nonoverlapping(left_starts.values, right_ends.values - 1, right_ends.index.values)

#     r_idx = pd.Series(r_idx, index=left_starts.index).sort_index().values
#     dist = pd.Series(dist, index=left_starts.index).sort_index().values

#     return r_idx, dist

# def sort_one_by_one(d, col1, col2):
#     """
#     Equivalent to pd.sort_values(by=[col1, col2]), but faster.
#     """
#     d = d.sort_values(by=[col2])
#     return d.sort_values(by=[col1], kind='mergesort')


# @return_empty_if_one_empty
# def _nearest(scdf, ocdf, suffix="_b", how=None, overlap=True, **kwargs):

#     if overlap:
#         nearest_df, df_to_find_nearest_in = _overlapping_for_nearest(scdf, ocdf, suffix, **kwargs)
#     else:
#         df_to_find_nearest_in = scdf

#     df_to_find_nearest_in = sort_one_by_one(df_to_find_nearest_in, "Start", "End")
#     ocdf = sort_one_by_one(ocdf, "Start", "End")
#     df_to_find_nearest_in.index = pd.Index(range(len(df_to_find_nearest_in)))

#     if how == "next":
#         r_idx, dist = _next_nonoverlapping(df_to_find_nearest_in.End, ocdf.Start, ocdf.index.values)
#     elif how == "previous":
#         r_idx, dist = _previous_nonoverlapping(df_to_find_nearest_in.Start, ocdf.End)
#     else:
#         previous_r_idx, previous_dist = _previous_nonoverlapping(df_to_find_nearest_in.Start, ocdf.End)

#         next_r_idx, next_dist = _next_nonoverlapping(df_to_find_nearest_in.End, ocdf.Start, ocdf.index.values)

#         r_idx, dist = nearest_nonoverlapping(previous_r_idx,
#                                              previous_dist,
#                                              next_r_idx, next_dist)

#     ocdf = ocdf.reindex(r_idx, fill_value=-1) # instead of np.nan, so ints are not promoted to float

#     ocdf.index = df_to_find_nearest_in.index
#     ocdf.insert(ocdf.shape[1], "Distance", pd.Series(dist, index=ocdf.index).fillna(-1).astype(int))

#     r_idx = pd.Series(r_idx, index=ocdf.index)
#     df_to_find_nearest_in = df_to_find_nearest_in.drop(r_idx.loc[r_idx == -1].index)

#     df = df_to_find_nearest_in.join(ocdf, rsuffix=suffix)

#     if overlap and not df.empty and not nearest_df.empty:
#         df = pd.concat([nearest_df, df])
#     elif overlap and not nearest_df.empty:
#         df = nearest_df

#     df = df.drop("Chromosome" + suffix, axis=1)
#     return df


# def _set_union(scdf, ocdf, **kwargs):

#     chromosome, strand = parse_grpby_key(kwargs["key"])

#     strandedness = kwargs["strandedness"]
#     strand = True if strandedness == "same" else False

#     if len(scdf) == 0:
#         return _cluster(ocdf, chromosome, strand=strand)
#     elif len(ocdf) == 0:
#         return _cluster(scdf, chromosome, strand=strand)

#     _starts = np.concatenate([
#         scdf.Start.values,
#         ocdf.Start.values])
#     _ends = np.concatenate([
#         scdf.End.values,
#         ocdf.End.values])

#     cdf = pd.DataFrame({"Start": _starts, "End": _ends})["Start End".split()]
#     cdf = cdf.sort_values("Start")
#     starts, ends = find_clusters(cdf.Start.values, cdf.End.values)

#     chromosome = scdf.head(1)["Chromosome"].iloc[0]
#     if strandedness == "same":
#         _strand = scdf.head(1)["Strand"].iloc[0]
#     else:
#         _strand = False

#     cluster_df = _create_df_from_starts_ends(starts, ends, chromosome, _strand)

#     return cluster_df


# def _subtraction(scdf, ocdf, **kwargs):

#     chromosome, strand = parse_grpby_key(kwargs["key"])

#     if ocdf.empty or scdf.empty:
#         return scdf

#     strandedness = kwargs["strandedness"]
#     strand = True if strandedness else False

#     oc = _cluster(ocdf, chromosome, strand)
#     o = NCLS(oc.Start.values, oc.End.values, oc.index.values)

#     idx_self, new_starts, new_ends = o.set_difference_helper(
#         scdf.Start.values,
#         scdf.End.values,
#         scdf.index.values)

#     missing_idx = pd.Index(scdf.index).difference(idx_self)

#     idx_to_drop = new_starts != -1

#     new_starts = new_starts[idx_to_drop]
#     new_ends = new_ends[idx_to_drop]

#     idx_self = idx_self[idx_to_drop]
#     new_starts = pd.Series(new_starts, index=idx_self).sort_index()
#     new_ends = pd.Series(new_ends, index=idx_self).sort_index()
#     idx_self = np.sort(idx_self)

#     scdf = scdf.reindex(missing_idx.union(idx_self))

#     if len(idx_self):

#         scdf.loc[scdf.index.isin(idx_self), "Start"] = new_starts
#         scdf.loc[scdf.index.isin(idx_self), "End"] = new_ends

#     return scdf


# def _coverage(ranges, value_col=None, stranded=False, n_jobs=1, **coverage):

#     try:
#         from pyranges import PyRles
#     except ImportError:
#         raise Exception("Using the coverage method requires that pyrle is installed.")

#     return PyRles(ranges, value_col=value_col, stranded=stranded, nb_cpu=n_jobs)



# @return_empty_if_one_empty
# def _write_both(scdf, ocdf, new_pos=False, **kwargs):

#     suffixes = kwargs["suffixes"]

#     scdf, ocdf = _both_dfs(scdf, ocdf, **kwargs)
#     nix = pd.Index(range(len(scdf)))
#     scdf.index = nix
#     ocdf.index = nix

#     ocdf = ocdf.drop("Chromosome", axis=1)

#     if not new_pos:
#         df = scdf.join(ocdf, rsuffix=suffixes[1])

#     elif new_pos == "intersection":

#         new_starts = pd.Series(
#             np.where(scdf.Start.values > ocdf.Start.values, scdf.Start, ocdf.Start),
#             index=scdf.index, dtype=np.long)

#         new_ends = pd.Series(
#             np.where(scdf.End.values < ocdf.End.values, scdf.End, ocdf.End),
#             index=scdf.index, dtype=np.long)
#         df = scdf.join(ocdf, lsuffix=suffixes[0], rsuffix=suffixes[1])
#         df.insert(1, "Start", new_starts)
#         df.insert(2, "End", new_ends)
#         df.rename(index=str, columns={"Chromosome" + suffixes[0]: "Chromosome", "Strand" + suffixes[0]: "Strand"}, inplace=True)

#     elif new_pos == "union":

#         new_starts = pd.Series(
#             np.where(scdf.Start.values < ocdf.Start.values, scdf.Start, ocdf.Start),
#             index=scdf.index, dtype=np.long)

#         new_ends = pd.Series(
#             np.where(scdf.End.values > ocdf.End.values, scdf.End, ocdf.End),
#             index=scdf.index, dtype=np.long)
#         df = scdf.join(ocdf, lsuffix=suffixes[0], rsuffix=suffixes[1])
#         df.insert(1, "Start", new_starts)
#         df.insert(2, "End", new_ends)
#         df.rename(index=str, columns={"Chromosome" + suffixes[0]: "Chromosome", "Strand" + suffixes[0]: "Strand"}, inplace=True)

#     return df


# if __name__ == "__main__":

#     chip_f = "/mnt/scratch/endrebak/pyranges_benchmark/data/download/chip_1000000.bed.gz"
#     background_f = "/mnt/scratch/endrebak/pyranges_benchmark/data/download/input_1000000.bed.gz"


#     chip = pd.read_table(chip_f, sep="\t", usecols=[0, 1, 2, 5], header=None,
#                             names="Chromosome Start End Strand".split(),
#                             dtype={"Chromosome": "category", "Strand": "category"} )

#     cgr = pr.PyRanges(chip, copy_df=False)

#     background = pd.read_table(background_f, sep="\t", usecols=[0, 1, 2, 5],
#                                 header=None, names="Chromosome Start End Strand".split(),
#                                 dtype={"Chromosome": "category", "Strand": "category"})

#     bgr = pr.PyRanges(background, copy_df=False)



import pytest

import pandas as pd

from pyranges import PyRanges


@pytest.fixture
def names():
    return "Chromosome  Start  End  Name Score Strand".split()

@pytest.fixture
def chip_10(names):

    df = pd.read_table("tests/chip_10.bed", header=None, names=names)

    gr = PyRanges(df)

    assert gr.stranded

    return gr

@pytest.fixture
def f1(names):

    df = pd.read_table("tests/f1.bed", sep="\t", header=None, names="Chromosome  Start  End  Name Score Strand".split())

    return PyRanges(df)


@pytest.fixture
def f2(names):

    df = pd.read_table("tests/f2.bed", sep="\t", header=None, names=names)

    return PyRanges(df)




import pytest

from hypothesis import given, settings, reproduce_failure, unlimited, HealthCheck, seed
from hypothesis.extra.pandas import data_frames, columns, range_indexes, column, indexes
from hypothesis.extra.numpy import arrays
import hypothesis.strategies as st

from itertools import product
import tempfile
import subprocess
from io import StringIO

from pyrle import Rle
import pyranges as pr

import pandas as pd
import numpy as np


from tests.helpers import assert_df_equal
from tests.hypothesis_helper import dfs_min



import numpy as np

from os import environ

if environ.get("TRAVIS"):
    max_examples = 100
    deadline = None
else:
    max_examples = 100
    deadline = None


merge_command = "bedtools merge -o first -c 6 {} -i <(sort -k1,1 -k2,2n {})"


@pytest.mark.bedtools
@pytest.mark.parametrize("strand", [True, False])
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min())
def test_cluster(gr, strand):

    bedtools_strand = {True: "-s", False: ""}[strand]

    print(gr)

    result_df = None
    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.bed".format(temp_dir)
        gr.df.to_csv(f1, sep="\t", header=False, index=False)

        cmd = merge_command.format(bedtools_strand, f1)
        print(cmd)

        result = subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()

        if not strand:
            bedtools_df = pd.read_table(StringIO(result), header=None, squeeze=True, usecols=[0, 1, 2], names="Chromosome Start End".split(), dtype={"Chromosome": "category"})
        else:
            bedtools_df = pd.read_table(StringIO(result), header=None, squeeze=True, names="Chromosome Start End Strand".split(), dtype={"Chromosome": "category"})

    print("bedtools_df\n", bedtools_df)
    result = gr.cluster(strand=strand)
    print("result\n", result.df)

    if not bedtools_df.empty:
        # need to sort because bedtools sometimes gives the result in non-natsorted chromosome order!
        if result.stranded:
            assert_df_equal(result.df.sort_values("Chromosome Start Strand".split()), bedtools_df.sort_values("Chromosome Start Strand".split()))
        else:
            assert_df_equal(result.df.sort_values("Chromosome Start".split()), bedtools_df.sort_values("Chromosome Start".split()))
    else:
        assert bedtools_df.empty == result.df.empty



from __future__ import print_function

import pandas as pd
# import numpy as np

import pkg_resources

from pyranges.pyranges import PyRanges

from pyrle import PyRles, Rle

# from pyranges.settings import pyranges_settings

# pyranges_settings = pyranges_settings

def load_dataset(basename):

    full_path = pkg_resources.resource_filename("pyranges", "example_data/{}.bed".format(basename))

    df = pd.read_table(full_path, header=None,
                       names="Chromosome Start End Name Score Strand".split())

    return PyRanges(df)


def list_datasets():

    datasets = [f.replace(".bed", "") for f in pkg_resources.resource_listdir("pyranges", "example_data")]

    print(datasets)


def read_bed(f):

    columns = "Chromosome Start End Name Score Strand ThickStart ThickEnd ItemRGB BlockCount BlockSizes BlockStarts".split()

    df = pd.read_table(f, dtype={"Chromosome": "category", "Strand": "category"}, header=None)

    df.columns = columns[:df.shape[1]]

    if df.Strand.str.contains("\.").any():
        print("Bed file contains '.' strands and is considered unstranded.", file=sys.stderr)

    return PyRanges(df)


def read_bam(f):

    import pysam

    samfile = pysam.AlignmentFile(f, "rb")

    chromosomes, starts, ends, names, sequences, strands = [], [], [], [], [], []
    for read in samfile.fetch():
        strand = '+'
        if(read.is_reverse):
            strand = '-'

        chromosomes.append(samfile.getrname(read.tid))
        starts.append(read.pos)
        ends.append(read.aend)
        sequences.append(read.seq)
        strands.append(strand)
        names.append(read.qname)

    df = pd.DataFrame({"Chromosome": chromosomes, "Start": starts, "End": ends,
                       "Sequence": sequences, "Strand": strands, "Name": names})["Chromosome Start End Sequence Name Strand".split()]

    return PyRanges(df)


def _fetch_gene_transcript_exon_id(attribute):

    no_quotes = attribute.str.replace('"', '').str.replace("'", "")

    df = no_quotes.str.extract("gene_id.?(.+?);(?:.*transcript_id.?(.+?);)?(?:.*exon_number.?(.+?);)?(?:.*exon_id.?(.+?);)?", expand=True)# .iloc[:, [1, 2, 3]]
    # print(df.head())
    # print(df.head().index)
    # print(df.columns)
    # raise

    df.columns = "GeneID TranscriptID ExonNumber ExonID".split()
    cat_cols = "GeneID TranscriptID ExonID".split()
    # df.columns = cat_cols
    df.loc[:, cat_cols] = df[cat_cols].astype("category")
    # print(df.head())

    return df


def read_gtf(f):

    """seqname - name of the chromosome or scaffold; chromosome names can be given with or without the 'chr' prefix. Important note: the seqname must be one used within Ensembl, i.e. a standard chromosome name or an Ensembl identifier such as a scaffold ID, without any additional content such as species or assembly. See the example GFF output below.
    # source - name of the program that generated this feature, or the data source (database or project name)
    feature - feature type name, e.g. Gene, Variation, Similarity
    start - Start position of the feature, with sequence numbering starting at 1.
    end - End position of the feature, with sequence numbering starting at 1.
    score - A floating point value.
    strand - defined as + (forward) or - (reverse).
    # frame - One of '0', '1' or '2'. '0' indicates that the first base of the feature is the first base of a codon, '1' that the second base is the first base of a codon, and so on..
    attribute - A semicolon-separated list of tag-value pairs, providing additional information about each feature."""
    dtypes = {"Chromosome": "category", "Feature": "category", "Strand": "category"}

    df = pd.read_table(f, sep="\t", comment="#", usecols=[0, 2, 3, 4, 5, 6, 8], header=None, names="Chromosome Feature Start End Score Strand Attribute".split(), dtype=dtypes)

    # Since Start is 1-indexed
    df.Start -= 1

    if sum(df.Score == ".") == len(df):
        cols_to_concat = "Chromosome Start End Strand Feature".split()
    else:
        cols_to_concat = "Chromosome Start End Strand Feature Score".split()

    extract = _fetch_gene_transcript_exon_id(df.Attribute)
    extract.columns = "GeneID TranscriptID ExonNumber ExonID".split()

    extract.ExonNumber = extract.ExonNumber.astype(float)

    df = pd.concat([df[cols_to_concat],
                        extract], axis=1)


    return PyRanges(df)


def get_example_path(basename):

    full_path = pkg_resources.resource_filename("pyranges", "example_data/{}".format(basename))

    if full_path.endswith(".bam"):
        _hack_to_load_idx = pkg_resources.resource_filename("pyranges", "example_data/{}.bai".format(basename))

    return full_path

read_gff = read_gtf





import pytest

from hypothesis import given, settings, reproduce_failure, unlimited, HealthCheck, seed
from hypothesis.extra.pandas import data_frames, columns, range_indexes, column, indexes
from hypothesis.extra.numpy import arrays
import hypothesis.strategies as st

from itertools import product
import tempfile
import subprocess
from io import StringIO

from pyrle import Rle
import pyranges as pr

import pandas as pd
import numpy as np


from tests.helpers import assert_df_equal
from tests.hypothesis_helper import dfs_min



import numpy as np

from os import environ

if environ.get("TRAVIS"):
    max_examples = 100
    deadline = None
else:
    max_examples = 100
    deadline = None


strandedness = [False, "same", "opposite"]
no_opposite = [False, "same"]


def run_bedtools(command, gr, gr2, strandedness, nearest_overlap=False):

    bedtools_strand = {False: "", "same": "-s", "opposite": "-S"}[strandedness]
    bedtools_overlap = {True: "", False: "-io"}[nearest_overlap]

    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.bed".format(temp_dir)
        f2 = "{}/f2.bed".format(temp_dir)
        gr.df.to_csv(f1, sep="\t", header=False, index=False)
        gr2.df.to_csv(f2, sep="\t", header=False, index=False)

        cmd = command.format(f1=f1, f2=f2, strand=bedtools_strand, overlap=bedtools_overlap)
        print(cmd)

        result = subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()

    return result


def read_bedtools_result_set_op(bedtools_result, strandedness):

    if strandedness:
        usecols = [0, 1, 2, 5]
        names = "Chromosome Start End Strand".split()
    else:
        usecols = [0, 1, 2]
        names = "Chromosome Start End".split()

    return pd.read_table(StringIO(bedtools_result), header=None, usecols=usecols, names=names)


def compare_results(bedtools_df, result):

    if not bedtools_df.empty:
        assert_df_equal(result.df, bedtools_df)
    else:
        assert bedtools_df.empty == result.df.empty


def compare_results_nearest(bedtools_df, result):

    print("1 " * 50)
    print(bedtools_df)
    if not bedtools_df.empty:
        bedtools_df = bedtools_df[bedtools_df.Distance != -1]

    print("2 " * 50)
    print(result)

    result = result.df

    if not len(result) == 0:
        result_df = result["Chromosome Start End Strand Distance".split()]
        assert_df_equal(result_df, bedtools_df)
    else:
        assert bedtools_df.empty


@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", no_opposite)
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())
def test_set_intersect(gr, gr2, strandedness):

    set_intersect_command = "bedtools intersect {strand} -a <(sort -k1,1 -k2,2n {f1} | bedtools merge {strand} -c 4,5,6 -o first -i -) -b <(sort -k1,1 -k2,2n {f2} | bedtools merge {strand} -c 4,5,6 -o first -i -)"
    bedtools_result = run_bedtools(set_intersect_command, gr, gr2, strandedness)

    bedtools_df = read_bedtools_result_set_op(bedtools_result, strandedness)

    result = gr.set_intersect(gr2, strandedness=strandedness)

    compare_results(bedtools_df, result)


@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", no_opposite)
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())
def test_set_union(gr, gr2, strandedness):

    set_union_command = "cat {f1} {f2} | bedtools sort | bedtools merge {strand} -c 4,5,6 -o first -i -"    # set_union_command = "bedtools merge {strand} -c 4,5,6 -o first -i {f1}"
    bedtools_result = run_bedtools(set_union_command, gr, gr2, strandedness)

    bedtools_df = read_bedtools_result_set_op(bedtools_result, strandedness)

    result = gr.set_union(gr2, strandedness=strandedness)

    compare_results(bedtools_df, result)


@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", strandedness)
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())
def test_overlap(gr, gr2, strandedness):

    overlap_command = "bedtools intersect -wa {strand} -a {f1} -b {f2}"

    bedtools_result = run_bedtools(overlap_command, gr, gr2, strandedness)

    bedtools_df = pd.read_table(StringIO(bedtools_result), header=None, names="Chromosome Start End Name Score Strand".split())

    result = gr.overlap(gr2, strandedness=strandedness)

    compare_results(bedtools_df, result)


@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", strandedness)
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())
def test_intersect(gr, gr2, strandedness):

    intersect_command = "bedtools intersect {strand} -a {f1} -b {f2}"

    bedtools_result = run_bedtools(intersect_command, gr, gr2, strandedness)

    bedtools_df = pd.read_table(StringIO(bedtools_result), header=None, names="Chromosome Start End Name Score Strand".split())

    result = gr.intersect(gr2, strandedness=strandedness)

    compare_results(bedtools_df, result)



@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", ["same", "opposite", False]) #
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())
def test_subtraction(gr, gr2, strandedness):

    subtract_command = "bedtools subtract {strand} -a {f1} -b {f2}"

    bedtools_result = run_bedtools(subtract_command, gr, gr2, strandedness)

    bedtools_df = pd.read_table(StringIO(bedtools_result), header=None, names="Chromosome Start End Name Score Strand".split())

    result = gr.subtract(gr2, strandedness=strandedness)

    compare_results(bedtools_df, result)


                    # "bedtools closest {} -t first -io -d -a <(sort -k1,1 -k2,2n {}) -b <(sort -k1,1 -k2,2n {})"]
nearest_hows = [None, "previous", "next"]
overlaps = [True, False]
strandedness = [False, "same", "opposite"]


@pytest.mark.bedtools
@pytest.mark.parametrize("nearest_how,overlap,strandedness", product(nearest_hows, overlaps, strandedness))
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())
def test_nearest(gr, gr2, nearest_how, overlap, strandedness):

    nearest_command = "bedtools closest {strand} {overlap} -t first -d -a <(sort -k1,1 -k2,2n {f1}) -b <(sort -k1,1 -k2,2n {f2})"

    bedtools_result = run_bedtools(nearest_command, gr, gr2, strandedness, overlap)

    bedtools_df = pd.read_table(StringIO(bedtools_result), header=None, names="Chromosome Start End Strand Distance".split(), usecols=[0, 1, 2, 5, 12])

    result = gr.nearest(gr2, strandedness=strandedness, overlap=overlap)

    compare_results_nearest(bedtools_df, result)



@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", no_opposite)
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())
def test_jaccard(gr, gr2, strandedness):

    # jaccard_command = "bedtools jaccard {strand}  -a <(sort -k1,1 -k2,2n {f1}) -b <(sort -k1,1 -k2,2n {f2})"

    # bedtools_result = run_bedtools(jaccard_command, gr, gr2, strandedness)

    # bedtools_jaccard = float(bedtools_result.split("\n")[1].split()[2])

    result = gr.jaccard(gr2, strandedness=strandedness)

    # there is a bug in bedtools, so cannot use as oracle
    assert 0 <= result <= 1
    # assert abs(result - bedtools_jaccard) < 0.001

# jaccard example, which should be 1, but bedtools considers 0.5:
# +--------------+-----------+-----------+------------+-----------+----------+
# | Chromosome   |     Start |       End | Name       |     Score | Strand   |
# | (int8)       |   (int32) |   (int32) | (object)   |   (int64) | (int8)   |
# |--------------+-----------+-----------+------------+-----------+----------|
# | chr1         |         1 |         2 | a          |         0 | +        |
# +--------------+-----------+-----------+------------+-----------+----------+
# PyRanges object has 1 sequences from 1 chromosomes.,
# =+--------------+-----------+-----------+------------+-----------+----------+
# | Chromosome   |     Start |       End | Name       |     Score | Strand   |
# | (int8)       |   (int32) |   (int32) | (object)   |   (int64) | (int8)   |
# |--------------+-----------+-----------+------------+-----------+----------|
# | chr1         |         1 |         2 | a          |         0 | +        |
# | chr1         |         1 |         2 | a          |         0 | -        |
# +--------------+-----------+-----------+------------+-----------+----------+
# PyRanges object has 2 sequences from 1 chromosomes., strandedness='same')
#
# bedtools jaccard -s  -a <(sort -k1,1 -k2,2n /tmp/tmpuxs7jtw0/f1.bed) -b <(sort -k1,1 -k2,2n /tmp/tmpuxs7jtw0/f2.bed)
# intersection	union-intersection	jaccard	n_intersections
# 1	2	0.5	1




@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", strandedness)
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())
def test_join(gr, gr2, strandedness):

    join_command = "bedtools intersect {strand} -wo -a {f1} -b {f2}"

    bedtools_result = run_bedtools(join_command, gr, gr2, strandedness)

    bedtools_df = pd.read_table(StringIO(bedtools_result), header=None,
                                names="Chromosome Start End Name Score Strand Chromosome_b Start_b End_b Name_b Score_b Strand_b Overlap".split(),
                                dtype={"Chromosome": "category", "Strand": "category"}).drop("Chromosome_b Overlap".split(), axis=1)

    result = gr.join(gr2, strandedness=strandedness)

    if result.df.empty:
        assert bedtools_df.empty
    else:
        assert_df_equal(result.df, bedtools_df)





import pytest
from tests.helpers import assert_df_equal

from pyranges.pyranges import PyRanges
import pyranges as pr

import pandas as pd

from io import StringIO


def test_tssify(f1):

    result = f1.tssify(slack=5)

    print(f1)
    print(result)

    assert list(result.Start) == [0, 3, 2]
    assert list(result.End) == [9, 14, 13]


def test_tesify(f2):

    print(f2)

    result = f2.tesify(slack=500)

    print(result)

    assert list(result.Start) == [0, 0]
    assert list(result.End) == [503, 508]



import pytest

from pyranges.pyranges import PyRanges
import pyranges as pr

import pandas as pd

from io import StringIO


def test_read_bam():

    pr.read_bam("tests/test_data/test_sorted.bam")


def test_read_gtf():

    gr = pr.read_gtf("tests/test_data/ensembl.gtf")

    assert list(gr.df.columns[:4]) == "Chromosome Start End Strand".split()


def test_read_gff3():

    gr = pr.read_gtf("tests/test_data/gencode.gff3")

    assert list(gr.df.columns[:4]) == "Chromosome Start End Strand".split()

def test_read_bed():

    pr.read_bed("pyranges/example_data/chipseq.bed")




from ncls import NCLS

def create_ncls(df):

    return NCLS(df.Start.values, df.End.values, df.index.values)


def find_overlaps(df, start, end):

    n = create_ncls(df)

    idxes = []
    for r in n.find_overlap(start, end):
        idxes.append(r[2])

    return idxes



import pandas as pd
from natsort import natsorted
import numpy as np
from collections import defaultdict, OrderedDict

import logging


try:
    # ray.init(logging_level=logging.CRITICAL) # logging_level=logging.CRITICAL # local_mode=True
    import ray
    ray.init(local_mode=True, logging_level=logging.CRITICAL) # logging_level=logging.CRITICAL # local_mode=True
except:
    import pyranges.raymock as ray

from tabulate import tabulate


# from pyranges.settings import pyranges_settings
from pyranges.genomicfeatures import GenomicFeaturesMethods
from pyranges.subset import get_string, get_slice, get_tuple
# from pyranges.methods import _cluster, _subtraction, _set_union, _set_intersection, _intersection, _nearest, _coverage, _overlap_write_both, _overlap, _tss, _tes, _jaccard, _lengths, _slack
from pyranges.multithreaded import (_cluster, pyrange_apply_single, _write_both, _jaccard, _coverage,
                                    _intersection, pyrange_apply, _nearest, _overlap, _first_df, _subtraction, _tss, _tes, _slack, _sort, merge_dfs, _concat)

def fill_kwargs(kwargs):

    if not "strandedness" in kwargs:
        kwargs["strandedness"] = "same"
    if not "suffix" in kwargs:
        kwargs["suffix"] = "_b"
    if not "overlap" in kwargs:
        kwargs["overlap"] = True
    if not "how" in kwargs:
        kwargs["how"] = None
    if not "invert" in kwargs:
        kwargs["invert"] = None
    if not "new_pos" in kwargs:
        kwargs["new_pos"] = None
    if not "suffixes" in kwargs:
        kwargs["suffixes"] = ["_a", "_b"]
    if not "suffix" in kwargs:
        kwargs["suffixes"] = ["_b"]

    return kwargs

def return_copy_if_view(df, df2):
    # https://stackoverflow.com/questions/26879073/checking-whether-data-frame-is-copy-or-view-in-pandas

    if df.values.base is df2.values.base:
        return df.copy(deep=True)
    else:
        return df


def create_df_dict(df):

    if "Strand" in df:
        grpby_key = "Chromosome Strand".split()
    else:
        grpby_key = "Chromosome"

    return {k: v for k, v in df.groupby(grpby_key)}


def create_pyranges_df(seqnames, starts, ends, strands=None):

    if isinstance(seqnames, str):
        seqnames = pd.Series([seqnames] * len(starts), dtype="category")

    if strands != None and strands != False:

        if isinstance(strands, str):
            strands = pd.Series([strands] * len(starts), dtype="category")

        columns = [seqnames, starts, ends, strands]
        lengths = list(str(len(s)) for s in columns)
        assert len(set(lengths)) == 1, "seqnames, starts, ends and strands must be of equal length. But are {}".format(", ".join(lengths))
        colnames = "Chromosome Start End Strand".split()
    else:
        columns = [seqnames, starts, ends]
        lengths = list(str(len(s)) for s in columns)
        assert len(set(lengths)) == 1, "seqnames, starts and ends must be of equal length. But are {}".format(", ".join(lengths))
        colnames = "Chromosome Start End".split()

    idx = range(len(starts))
    series_to_concat = []
    for s in columns:
        s = pd.Series(s, index=idx)
        series_to_concat.append(s)

    df = pd.concat(series_to_concat, axis=1)
    df.columns = colnames

    return df


def return_empty_if_one_empty(func):

    def extended_func(self, other, *args, **kwargs):

        if len(self) == 0 or len(other) == 0:
            df = pd.DataFrame(columns="Chromosome Start End Strand".split())
        else:
            df = func(self, other, *args, **kwargs)

        return df

    return extended_func


def return_empty_if_both_empty(func):

    def extended_func(self, other, *args, **kwargs):

        if len(self) == 0 and len(other) == 0:
            df = pd.DataFrame(columns="Chromosome Start End Strand".split())
        else:
            df = func(self, other, *args, **kwargs)

        return df

    return extended_func


def pyrange_or_df(func):

    def extension(self, other, *args, **kwargs):
        df = func(self, other, *args, **kwargs)

        if kwargs.get("df"):
            return df

        return PyRanges(df)

    return extension


def pyrange_or_df_single(func):

    def extension(self, *args, **kwargs):
        df = func(self, *args, **kwargs)

        if kwargs.get("df"):
            return df

        return PyRanges(df)

    return extension




def set_dtypes(df, extended):

    if not extended:
        dtypes = {"Start": np.int32, "End": np.int32, "Chromosome": "category", "Strand": "category"}
    else:
        dtypes = {"Start": np.int64, "End": np.int64, "Chromosome": "category", "Strand": "category"}

    if not "Strand" in df:
        del dtypes["Strand"]


    if "ExonNumber" in df:
        df.ExonNumber = df.ExonNumber.fillna(-1)
        if df.ExonNumber.max() < 128:
            df.ExonNumber = df.ExonNumber.astype(np.int8)
        else:
            df.ExonNumber = df.ExonNumber.astype(np.int16)

    for col, dtype in dtypes.items():

        if df[col].dtype.name != dtype:
            df[col] = df[col].astype(dtype)

    return df


class PyRanges():

    dfs = None
    gf = None

    def __init__(self, df=None, seqnames=None, starts=None, ends=None, strands=None, copy_df=False, extended=False):

        if isinstance(df, PyRanges):
            raise Exception("Object is already a PyRange.")


        if copy_df:
            df = df.copy()

        if isinstance(df, pd.DataFrame):
            df = set_dtypes(df, extended)

        if df is False or df is None:
            df = create_pyranges_df(seqnames, starts, ends, strands)
            self.__dict__["dfs"] = create_df_dict(df)
        elif isinstance(df, pd.DataFrame):
            self.__dict__["dfs"] = create_df_dict(df)
        else:
            # df is actually dict of dfs
            self.__dict__["dfs"] = df

        self.__dict__["ft"] = GenomicFeaturesMethods(self)


    def __len__(self):
        return sum([ len(d) for d in self.values() ])


    def __getattr__(self, name):

        if name in self.values()[0]:
            return pd.concat([df[name] for df in self.values()])
        else:
            raise Exception("PyRanges object has no attribute", name)


    def __setattr__(self, column_name, column):

        if column_name in "Chromosome Strand".split():
            raise Exception("The columns Chromosome and Strand can not be reset.")

        isiterable = isinstance(column, list) or isinstance(column, pd.Series) or isinstance(column, np.ndarray)
        if isiterable:
            if not len(self) == len(column):
                raise Exception("DataFrame and column must be same length.")

        already_exists = column_name in self.values()[0]

        if already_exists:
            pos = list(self.values()[0].columns).index(column_name)
        else:
            pos = self.values()[0].shape[1]

        start_length, end_length = 0, 0

        dfs = {}
        for k, df in self.items():

            end_length += len(df)

            if already_exists:
                df = df.drop(column_name, axis=1)

            if isiterable:
                df.insert(pos, column_name, column[start_length:end_length])
            else:
                df.insert(pos, column_name, column)

            start_length = end_length

            dfs[k] = df

        self.__dict__["dfs"] = dfs


    def __eq__(self, other):

        return self.df.equals(other.df)

    def __getitem__(self, val):

        # print("self, val")
        # print(self, val)
        if isinstance(val, str):
            df = get_string(self, val)
        elif isinstance(val, tuple):
            df = get_tuple(self, val)
        elif isinstance(val, slice):
            df = get_slice(self, val)
        else:
            raise Exception("Not valid subsetter: {}".format(str(val)))

        # print("getitem " * 100)
        # print(df)
        # print(ray.get(df))
        return PyRanges(df)


    def __str__(self):

        # print("in str")
        if len(self) == 0:
            return "Empty PyRanges"

        # keys = natsorted(list(self.dfs.keys()))
        if len(self.keys()) == 1:

            first_key = self.keys()[0]
            # last_key = list(self.dfs.keys())[-1]
            first_df = self.dfs[first_key]

            # last_df = ray.get(self.dfs[last_key]).tail(3)
            h = first_df.head(3).astype(object)
            m = first_df.head(1).astype(object)
            t = first_df.tail(3).astype(object)
            m.loc[:,:] = "..."

            if len(self) > 6:
                s = pd.concat([h, m, t])
            elif len(self) == 6:
                s = pd.concat([h, t])
            else:
                s = h
        else:
            keys = self.keys()
            first_key = keys[0]
            last_key = keys[-1]
            # first_key = self.keys[0]
            # last_key = self.keys[-1]
            first_df = self.dfs[first_key].head(3)
            last_df = self.dfs[last_key].tail(3)
            # last_df = self.dfs[list(self.dfs.keys())[-1]].tail(3)

            h = first_df.head(3).astype(object)
            m = first_df.head(1).astype(object)
            t = last_df.head(3).astype(object)
            m.loc[:,:] = "..."
            # m.index = ["..."]
            # print((len(h) + len(t)) < 6, len(self) >= 6)
            if (len(h) + len(t)) < 6:

                keys_covered = set()
                # iterate from front until have three
                heads = []
                hl = 0
                for k in keys:
                    keys_covered.add(k)
                    h = self.dfs[k].head(3)
                    first_df = h
                    hl += len(h)
                    heads.append(h)
                    if hl >= 3:
                        break

                tails = []
                tl = 0
                for k in keys[::-1]:
                    if k in keys_covered:
                        continue

                    t = self.dfs[k].tail(3)
                    tl += len(t)
                    tails.append(t)
                    if tl >= 3:
                        break
                # iterate from back until have three

                h = pd.concat(heads).head(3).astype(object)
                if tails:
                    t = pd.concat(tails).tail(3).astype(object)
                    if len(h) + len(t) > 6:
                        m = h.head(1).astype(object)
                        m.loc[:,:] = "..."
                        s = pd.concat([h, m, t])
                    else:
                        s = pd.concat([h, t])
                else:
                    s = h

            elif len(h) + len(t) == 6:
                m.loc[:,:] = "..."
                s = pd.concat([h, m, t])
            else:
                s = pd.concat([h, t])

        if False: # make setting
            if self.stranded:
                pos = s.Chromosome.astype(str) + " " + s.Start.astype(str) + "-" + s.End.astype(str) + " " + s.Strand.astype(str)
                s = s.drop("Chromosome Start End Strand".split(), axis=1)
                first_df = first_df.drop("Chromosome Start End Strand".split(), axis=1)
            else:
                pos = s.Chromosome.astype(str) + " " + s.Start.astype(str) + "-" + s.End.astype(str)
                s = s.drop("Chromosome Start End".split(), axis=1)
                first_df = first_df.drop("Chromosome Start End".split(), axis=1)

            s.insert(0, "Position", pos)
            h = [c + "\n(" + str(t) + ")" for c, t in  zip(s.columns, ["multiple types"] + list(first_df.dtypes))]
        else:
            dtypes = []
            for col, dtype in zip(s.columns, first_df.dtypes):
                if str(dtype) == "category":
                    dtype = first_df[col].cat.codes.dtype

                # dtype = str(dtype).replace("float", "f_").replace("int", "i_")
                dtypes.append(dtype)

            h = [c + "\n(" + str(t) + ")" for c, t in  zip(s.columns, list(dtypes))]

        str_repr = tabulate(s, headers=h, tablefmt='psql', showindex=False) + \
                                        "\nPyRanges object has {} sequences from {} chromosomes.".format(len(self), len(self.chromosomes))
        return str_repr


    def __repr__(self):

        return str(self)


    # @pyrange_or_df
    # @return_empty_if_one_empty
    def overlap(self, other, **kwargs):

        "Want all intervals in self that overlap with other."

        # print(kwargs)
        kwargs = fill_kwargs(kwargs)
        # print(kwargs)

        dfs = pyrange_apply(_overlap, self, other, **kwargs)

        # df = _overlap(self, other, strandedness, invert, how)

        return PyRanges(dfs)

    # @pyrange_or_df
    # @return_empty_if_one_empty
    def nearest(self, other, **kwargs):

        "Find the nearest feature in other."

        kwargs = fill_kwargs(kwargs)

        dfs = pyrange_apply(_nearest, self, other, **kwargs)

        return PyRanges(dfs)

    # @return_empty_if_one_empty
    def intersect(self, other, **kwargs):

        kwargs = fill_kwargs(kwargs)

        dfs = pyrange_apply(_intersection, self, other, **kwargs)

        return PyRanges(dfs)

    # @return_empty_if_one_empty
    def set_intersect(self, other, **kwargs):

        kwargs = fill_kwargs(kwargs)
        strandedness = kwargs["strandedness"]
        strand = True if strandedness else False
        self_clusters = self.cluster(strand=strand, **kwargs)
        other_clusters = other.cluster(strand=strand, **kwargs)
        dfs = pyrange_apply(_intersection, self_clusters, other_clusters, **kwargs)

        return PyRanges(dfs)

    def set_union(self, other, **kwargs):

        kwargs = fill_kwargs(kwargs)
        strandedness = kwargs["strandedness"]
        strand = True if strandedness else False

        pr = self.concat(other)
        # print(dfs)
        # print(pr)
        pr = pr.cluster(strand=strand, **kwargs)

        return pr


    # @pyrange_or_df
    def subtract(self, other, **kwargs):

        kwargs = fill_kwargs(kwargs)

        result = pyrange_apply(_subtraction, self, other, **kwargs)

        return PyRanges(result)


    # @pyrange_or_df
    # @return_empty_if_one_empty
    def join(self, other, **kwargs):

        kwargs = fill_kwargs(kwargs)
        dfs = pyrange_apply(_write_both, self, other, **kwargs)

        return PyRanges(dfs)


    def cluster(self, strand=None, **kwargs):

        df = pyrange_apply_single(_cluster, self, strand, kwargs)

        return PyRanges(df)


    def coverage(self, value_col=None, strand=True):

        return _coverage(self, value_col, strand=strand)


    def apply(self, f, strand=True, as_pyranges=True, kwargs=None):

        if kwargs is None:
            kwargs = {}

        f = ray.remote(f)

        result = pyrange_apply_single(f, self, strand, kwargs)

        if not as_pyranges:
            return result
        else:
            return PyRanges(result)


    def apply_pair(self, other, f, kwargs, strand=True, as_pyranges=True):

        f = ray.remote(f)

        result = pyrange_apply(f, self, other, strand, kwargs)

        if not as_pyranges:
            return result
        else:
            return PyRanges(result)

    def concat(self, other):

        return PyRanges(_concat(self, other))



    def jaccard(self, other, **kwargs):

        kwargs = fill_kwargs(kwargs)
        strand = True if kwargs["strandedness"] else False
        self_clusters = self.cluster(strand=strand)
        other_clusters = other.cluster(strand=strand)

        results = pyrange_apply(_jaccard, self_clusters, other_clusters, **kwargs)
        values = results.values()

        ssum, osum, ilsum = 0, 0, 0
        for s, o, il in values:
            # print(s, o, il)
            ssum += s
            osum += o
            ilsum += il

        if ssum + osum == ilsum:
            jaccard = 1
        else:
            jaccard = ilsum / (ssum + osum - ilsum)

        return jaccard

    def slack(self, slack):

        kwargs = {"slack": slack}
        prg = PyRanges(pyrange_apply_single(_slack, self, self.stranded, kwargs))

        return prg

    def tssify(self, slack=0):

        kwargs = {"slack": slack}
        return PyRanges(pyrange_apply_single(_tss, self, self.stranded, kwargs))

    def sort(self, columns=["Start", "End"], **kwargs):

        kwargs = fill_kwargs(kwargs)
        return PyRanges(pyrange_apply_single(_sort, self, self.stranded, kwargs))

    def tesify(self, slack=0):

        kwargs = {"slack": slack}
        return PyRanges(pyrange_apply_single(_tes, self, self.stranded, kwargs))

    def pos(self, *val):

        # turn int-tuple into slice
        newval = []
        for v in val:
            if isinstance(v, tuple) and len(v) == 2 and isinstance(v[0], int) and isinstance(v[1], int):
                newval.append(slice(v[0], v[1]))
            else:
                newval.append(v)

        val = tuple(newval)

        if len(val) == 1:
            val = val[0]

        if isinstance(val, str):
            df = get_string(self, val)
        elif isinstance(val, tuple):
            df = get_tuple(self, val)
        elif isinstance(val, slice):
            df = get_slice(self, val)
        else:
            raise Exception("Invalid type for indexer. Must be str, slice, or 2/3-tuple.")

        return return_copy_if_view(df, self.df)


    # always returns df
    def get(self, column, values, **kwargs):

        df = self.df
        if not column in self.df:
            raise Exception("Column {} not in PyRanges".format(column))

        if isinstance(values, str):
            df = df.loc[df[column] == values]
        else:
            df = df.loc[df[column].isin(values)]

        return return_copy_if_view(df, self.df)


    # @property
    def keys(self):
        return natsorted(self.dfs.keys())

    @property
    def stranded(self):
        return len(list(self.keys())[0]) == 2

    @property
    def strands(self):

        if not self.stranded:
            raise Exception("PyRanges not stranded!")

        return natsorted(set([k[1] for k in self.keys()]))

    @property
    def chromosomes(self):

        if self.stranded:
            return natsorted(set([k[0] for k in self.keys()]))
        else:
            return natsorted(set([k for k in self.keys()]))


    # @property
    def items(self):

        return natsorted([(k, df) for (k, df) in self.dfs.items()])

    # @property
    def values(self):

        return [df for k, df in self.items() if not df.empty]

    # @property
    # def objids(self):

    #     return [objid for k, objid in natsorted(self.dfs.items())]

    @property
    def df(self):

        return self.as_df()

    def as_df(self):

        if len(self) == 0:
            return pd.DataFrame()
        elif len(self) == 1:
            return self.values()[0]
        else:
            return pd.concat(self.values())


    def lengths(self):

        lengths = {}
        for k, df in self.items():
            lengths[k] = df.End - df.Start

        return lengths


    def summary(self, chromosome=False, strand=False):

        lengths = OrderedDict()
        lengths["pyrange"] = self.lengths()

        if self.stranded:
            c = self.cluster(strand=True)
            lengths["coverage_stranded"] = c.lengths()

        c = self.cluster(strand=False)
        lengths["coverage_unstranded"] = c.lengths()

        summaries = OrderedDict()
        # if not chromosome and not strand:

        for summary, d in lengths.items():
            summaries[summary] = pd.concat(d.values()).describe()
        # else:
        #     raise Exception("Flag chromosome and strand not implemented yet!")


        summary = pd.concat(summaries.values(), axis=1)
        summary.columns = list(summaries)


        str_repr = tabulate(summary, headers=summary.columns, tablefmt='psql')
        print(str_repr)

            # pd.concat()



import os
import sys

from distutils.core import setup
from setuptools import find_packages

install_requires = ["cython", "pandas", "ncls", "tabulate", "pysam", "sorted_nearest", "pyrle"]


setup(
    name="pyranges",
    packages=find_packages(),
    package_data={'pyranges': ['example_data/*.bed', 'example_data/*.gtf',
                               'example_data/*.bam', 'example_data/*.bam.bai']},
    include_dirs=["."],
    version="0.0.15",
    description="GenomicRanges for Python.",
    author="Endre Bakken Stovner",
    author_email="endrebak85@gmail.com",
    url="http://github.com/endrebak/pyranges",
    keywords=["Bioinformatics"],
    license=["MIT"],
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment", "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        'License :: OSI Approved :: MIT License',
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Topic :: Scientific/Engineering"
    ],
    long_description=
    ("Performant Pythonic GenomicRanges."))



import numpy as np
import pandas as pd
from ncls import NCLS

import pyranges as pr

from natsort import natsorted

from sorted_nearest import (find_clusters, nearest_previous_nonoverlapping,
                            nearest_next_nonoverlapping, nearest_nonoverlapping, find_clusters)

from collections import defaultdict

from functools import wraps

try:
    import ray
except:
    import pyranges.raymock as ray

import sys

@ray.remote
def _create_df_from_starts_ends(starts, ends, chromosome, strand=None):

    if not cluster_df.empty:
        return cluster_df
    else:
        return None


#     if isinstance(grpby_key, str):
#         return grpby_key, False
#     else:
#         return grpby_key[0], grpby_key[1]


def return_empty_if_one_empty(func):

    @wraps(func)
    def extended_func(self, other, **kwargs):

        if len(self) == 0 or len(other) == 0:
            df = pd.DataFrame(columns="Chromosome Start End".split())
        else:
            df = func(self, other, **kwargs)

        return df

    return extended_func


@ray.remote
def merge_dfs(df1, df2, kwargs):

    if not df1.empty and not df2.empty:
        return pd.concat([df1, df2], sort=False)

    elif df1.empty and df2.empty:
        # can this happen?
        return None
    elif df1.empty:
        return df2
    else:
        return df1


def _concat(self, other):

    strand = False

    if self.stranded and not other.stranded:
        self = pr.PyRanges(pyrange_apply_single(merge_dfs, self, strand))
    elif not self.stranded and other.stranded:
        other = pr.PyRanges(pyrange_apply_single(merge_dfs, other, strand))

    keys = natsorted(set(self.keys() + other.keys()))

    concatted = []
    for k in keys:
        s = self.dfs.get(k, pd.DataFrame(columns="Chromosome Start End".split()))
        o = other.dfs.get(k, pd.DataFrame(columns="Chromosome Start End".split()))

        result = merge_dfs.remote(s, o, {})
        concatted.append(result)

    results = ray.get(concatted)

    return {k: v for (k, v) in zip(keys, results) if not v is None}


def pyrange_apply(function, self, other, **kwargs):

    strandedness = kwargs["strandedness"]

    other_strand = {"+": "-", "-": "+"}
    same_strand = {"+": "+", "-": "-"}

    if strandedness == "opposite":
        strand_dict = other_strand
    else:
        strand_dict = same_strand

    assert strandedness in ["same", "opposite", False, None]

    if strandedness:
        assert self.stranded and other.stranded, \
            "Can only do stranded operations when both PyRanges contain strand info"

    results = []

    items = natsorted(self.dfs.items())
    keys = natsorted(self.dfs.keys())


    if strandedness:

        for (c, s), df in items:

            os = strand_dict[s]
            # print("s", s, "os", os)
            # print(self.keys)
            # print(other.keys)

            # cannot do this with set subtraction
            # but need it for ...
            if not (c, os) in other.keys():
                # print(c, os, "not in ", other.keys)
                odf = pd.DataFrame(columns="Chromosome Start End".split())
            else:
                # print(other)
                odf = other[c, os].values()[0]


            # print(c, s)
            # print(df.head())
            # print(odf.head())
            # print(df.dtypes)
            # print(odf.dtypes)
            # print("other", other)
            # print("other[c, os]", other[c, os])
            # odf = other[c, os].values[0]
            result = function.remote(df, odf, kwargs)
            # print("successfully completed")
            # print(result)
            # print(result)
            # print(" --- " * 50)
            # print(result)
            results.append(result)

    else:

        if self.stranded and not other.stranded:

            for (c, s), df in items:

                if not c in other.chromosomes:
                    odf = pr.PyRanges(pd.DataFrame(columns="Chromosome Start End".split()))
                else:
                    odf = other.dfs[c]

                result = function.remote(df, odf, kwargs)
                results.append(result)


        elif not self.stranded and other.stranded:

            for c, df in items:

                if not c in other.chromosomes:
                    odf = pr.PyRanges(pd.DataFrame(columns="Chromosome Start End".split()))
                else:
                    odf1 = other[c, "+"]
                    odf2 = other[c, "-"]
                    # merge strands
                    odf = merge_dfs.remote(odf1, odf2)

                result = function.remote(df, odf, kwargs)
                results.append(result)

        elif self.stranded and other.stranded:

            for (c, s), df in self.items():

                if not c in other.chromosomes:
                    odfs = pr.PyRanges(pd.DataFrame(columns="Chromosome Start End".split()))
                else:
                    odfs = other[c].values()


                if len(odfs) == 2:
                    odf = merge_dfs.remote(*odfs, kwargs)
                elif len(odfs) == 1:
                    odf = odfs[0]
                else:
                    odf = pd.DataFrame(columns="Chromosome Start End".split())


                result = function.remote(df, odf, kwargs)
                results.append(result)

        else:

            for c, df in items:
                if not c in other.chromosomes:
                    odf = pd.DataFrame(columns="Chromosome Start End".split())
                else:
                    odf = other.dfs[c]

                result = function.remote(df, odf, kwargs)
                results.append(result)

    results = ray.get(results)


    return {k: r for k, r in zip(keys, results) if r is not None}



def pyrange_apply_single(function, self, strand, kwargs):

    if strand:
        assert self.stranded, \
            "Can only do stranded operation when PyRange contains strand info"

    if self.stranded and strand:
        grpby_key = ["Chromosome", "Strand"]
    else:
        grpby_key = "Chromosome"

    items = self.items()

    results = []

    if strand:

        for (c, s), df in items:

            kwargs["chromosome"] = c
            _strand = s
            kwargs["strand"] = _strand
            result = function.remote(df, kwargs)
            results.append(result)

        keys = self.keys()
        # raise

    elif not self.stranded:

        keys = []
        for c, df in items:

            kwargs["chromosome"] = c
            result = function.remote(df, kwargs)
            results.append(result)
            keys.append(c)

    else:

        keys = []
        for c in self.chromosomes:

            kwargs["chromosome"] = c

            dfs = self[c]
            # print(dfs.values)
            # print(type(dfs.values))
            # print(len(dfs.values))
            if len(dfs.keys()) == 2:
                df1, df2 = dfs.values()
                # merge strands
                df1 = merge_dfs.remote(df1, df2, kwargs)
            else:
                df1 = dfs.values()[0]
            # print(type( df1 ))
            # print(type( df2 ))
            # print(df1)
            # print(df2)
            result = function.remote(df1,kwargs)
            results.append(result)
            keys.append(c)

    results = ray.get(results)

    return {k: r for k, r in zip(keys, results) if r is not None}


@ray.remote
def _cluster(df, kwargs):

    chromosome, strand = kwargs["chromosome"], kwargs.get("strand", None)

    cdf = df.sort_values("Start")

    # print("ooooooool" * 10)
    # cdf = sort_one_by_one(df, "Start", "End")

    starts, ends = find_clusters(cdf.Start.values, cdf.End.values)

    nidx = pd.Index(range(len(starts)))
    if strand:
        cluster_df = pd.DataFrame({"Chromosome": pd.Series(chromosome, dtype="category", index=nidx),
                                   "Start": starts, "End": ends,
                                   "Strand": pd.Series(strand, dtype="category", index=nidx)})
    else:
        cluster_df = pd.DataFrame({"Chromosome": pd.Series(chromosome, dtype="category", index=nidx), "Start": starts, "End": ends})

    return cluster_df


@ray.remote
def _first_df(scdf, ocdf, kwargs):

    if scdf.empty or ocdf.empty:
        return None

    how = kwargs["how"]

    assert how in "containment first".split() + [False, None]
    starts = scdf.Start.values
    ends = scdf.End.values
    indexes = scdf.index.values

    it = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

    if not how:
        _indexes = it.has_overlaps(starts, ends, indexes)
    elif how == "containment":
        _indexes = it.has_containment(starts, ends, indexes)

    return scdf.reindex(_indexes)

@ray.remote
def _overlap(scdf, ocdf, kwargs):

    if scdf.empty or ocdf.empty:
        return None

    how = kwargs["how"]

    assert how in "containment first".split() + [False, None]
    starts = scdf.Start.values
    ends = scdf.End.values
    indexes = scdf.index.values

    it = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

    _indexes = it.all_overlaps_self(starts, ends, indexes)

    return scdf.reindex(_indexes)


def _both_dfs(scdf, ocdf, how=False):

    assert how in "containment first".split() + [False, None]
    starts = scdf.Start.values
    ends = scdf.End.values
    indexes = scdf.index.values


    it = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

    if not how:
        _self_indexes, _other_indexes = it.all_overlaps_both(starts, ends, indexes)
    elif how == "containment":
        _self_indexes, _other_indexes = it.all_containments_both(starts, ends, indexes)
    else:
        _self_indexes, _other_indexes = it.first_overlap_both(starts, ends, indexes)

    _self_indexes = _self_indexes
    _other_indexes = _other_indexes

    return scdf.reindex(_self_indexes), ocdf.reindex(_other_indexes)

@ray.remote
def _intersection(scdf, ocdf, kwargs):

    how = kwargs["how"]

    if ocdf.empty or scdf.empty: # just return empty df
        return None
    # scdf2, ocdfe#2 = ray.get(_both_dfs.remote(scdf, ocdf, how=False))
    assert how in "containment first".split() + [False, None]
    starts = scdf.Start.values
    ends = scdf.End.values
    indexes = scdf.index.values

    in_dtype = ocdf.Start.dtype

    oncls = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

    if not how or how is None:
        _self_indexes, _other_indexes = oncls.all_overlaps_both(starts, ends, indexes)
    elif how == "containment":
        _self_indexes, _other_indexes = oncls.all_containments_both(starts, ends, indexes)
    elif how == "both":
        _self_indexes, _other_indexes = oncls.first_overlap_both(starts, ends, indexes)

    _self_indexes = _self_indexes
    _other_indexes = _other_indexes

    scdf, ocdf = scdf.reindex(_self_indexes), ocdf.reindex(_other_indexes)


    new_starts = pd.Series(
        np.where(scdf.Start.values > ocdf.Start.values, scdf.Start, ocdf.Start),
        index=scdf.index, dtype=in_dtype)

    new_ends = pd.Series(
        np.where(scdf.End.values < ocdf.End.values, scdf.End, ocdf.End),
        index=scdf.index, dtype=in_dtype)

    pd.options.mode.chained_assignment = None  # default='warn'
    scdf.loc[:, "Start"] = new_starts
    scdf.loc[:, "End"] = new_ends
    pd.options.mode.chained_assignment = 'warn'

    if not scdf.empty:
        return scdf
    else:
        return None





# @ray.remote
# def _set_intersection(scdf, ocdf, kwargs):

#     if len(scdf) == 0 or len(ocdf) == 0:
#         return None

#     chromosome = scdf.Chromosome.iloc[0]

#     strand = True if kwargs["strandedness"] else False
#     if strand:
#         strand = scdf.Strand.iloc[0]

#     s = _cluster.remote(scdf, chromosome, strand=strand)
#     o = _cluster.remote(ocdf, chromosome, strand=strand)

#     return _intersection.remote(s, o, kwargs)



def sort_one_by_one(d, col1, col2):
    """
    Equivalent to pd.sort_values(by=[col1, col2]), but faster.
    """
    d = d.sort_values(by=[col2])
    return d.sort_values(by=[col1], kind='mergesort')

def _overlapping_for_nearest(scdf, ocdf, suffix):

    nearest_df = pd.DataFrame(columns="Chromosome Start End Strand".split())

    scdf2, ocdf2 = _both_dfs(scdf, ocdf, how="first")


    if not ocdf2.empty:
        # only copying data because of the eternal source buffer array is read only problem
        original_idx = scdf.index

        idxs = scdf2.index
        original_idx = scdf.index.copy(deep=True)
        missing_idxs = ~original_idx.isin(idxs)
        missing_overlap = scdf.index[missing_idxs]

        df_to_find_nearest_in = scdf.reindex(missing_overlap)

        odf = ocdf.reindex(ocdf2.index)
        odf.index = idxs
        sdf = scdf.reindex(idxs)

        nearest_df = sdf.join(odf, rsuffix=suffix)
        nearest_df.insert(nearest_df.shape[1], "Distance", 0)
    else:
        df_to_find_nearest_in = scdf

    return nearest_df, df_to_find_nearest_in


def _next_nonoverlapping(left_ends, right_starts, right_indexes):

    left_ends = left_ends.sort_values()
    right_starts = right_starts.sort_values()
    r_idx, dist = nearest_next_nonoverlapping(left_ends.values - 1, right_starts.values, right_indexes)
    r_idx = pd.Series(r_idx, index=left_ends.index).sort_index().values
    dist = pd.Series(dist, index=left_ends.index).sort_index().values

    return r_idx, dist


def _previous_nonoverlapping(left_starts, right_ends):

    left_starts = left_starts.sort_values()
    right_ends = right_ends.sort_values()
    r_idx, dist = nearest_previous_nonoverlapping(left_starts.values, right_ends.values - 1, right_ends.index.values)

    r_idx = pd.Series(r_idx, index=left_starts.index).sort_index().values
    dist = pd.Series(dist, index=left_starts.index).sort_index().values

    return r_idx, dist



@ray.remote
def _nearest(scdf, ocdf, kwargs):


    if scdf.empty or ocdf.empty:
        return None

    overlap = kwargs["overlap"]
    how = kwargs["how"]
    suffix = kwargs["suffix"]

    if overlap:
        nearest_df, df_to_find_nearest_in = _overlapping_for_nearest(scdf, ocdf, suffix)
    else:
        df_to_find_nearest_in = scdf

    if not df_to_find_nearest_in.empty:
        df_to_find_nearest_in = sort_one_by_one(df_to_find_nearest_in, "Start", "End")
        ocdf = sort_one_by_one(ocdf, "Start", "End")
        df_to_find_nearest_in.index = pd.Index(range(len(df_to_find_nearest_in)))

        if how == "next":
            r_idx, dist = _next_nonoverlapping(df_to_find_nearest_in.End, ocdf.Start, ocdf.index.values)
        elif how == "previous":
            r_idx, dist = _previous_nonoverlapping(df_to_find_nearest_in.Start, ocdf.End)
        else:
            previous_r_idx, previous_dist = _previous_nonoverlapping(df_to_find_nearest_in.Start, ocdf.End)

            next_r_idx, next_dist = _next_nonoverlapping(df_to_find_nearest_in.End, ocdf.Start, ocdf.index.values)

            r_idx, dist = nearest_nonoverlapping(previous_r_idx,
                                                previous_dist,
                                                next_r_idx, next_dist)


        # sys.stderr.flush()

        ocdf = ocdf.reindex(r_idx, fill_value=-1) # instead of np.nan, so ints are not promoted to float

        ocdf.index = df_to_find_nearest_in.index
        ocdf.insert(ocdf.shape[1], "Distance", pd.Series(dist, index=ocdf.index).fillna(-1).astype(int))

        r_idx = pd.Series(r_idx, index=ocdf.index)
        df_to_find_nearest_in = df_to_find_nearest_in.drop(r_idx.loc[r_idx == -1].index)

        df = df_to_find_nearest_in.join(ocdf, rsuffix=suffix)

    if overlap and "df" in locals() and not df.empty and not nearest_df.empty:
        df = pd.concat([nearest_df, df])
    elif overlap and not nearest_df.empty:
        df = nearest_df

    df = df.drop("Chromosome" + suffix, axis=1)
    return df



# @ray.remote
# def _nearest(scdf, ocdf, kwargs):

#     if scdf.empty or ocdf.empty:
#         return None

#     # suffix="_b", how=None, overlap=True
#     overlap = kwargs["overlap"]
#     how = kwargs["how"]
#     suffix = kwargs["suffix"]
#     strandedness = kwargs["strandedness"]

#     if overlap:
#         nearest_df, df_to_find_nearest_in = _overlapping_for_nearest(scdf, ocdf, strandedness, suffix)
#     else:
#         df_to_find_nearest_in = scdf


#     if not df_to_find_nearest_in.empty:
#         df_to_find_nearest_in = sort_one_by_one(df_to_find_nearest_in, "Start", "End")
#         ocdf = sort_one_by_one(ocdf, "Start", "End")
#         df_to_find_nearest_in.index = pd.Index(range(len(df_to_find_nearest_in)))

#         if how == "next":
#             l_idx, r_idx, dist = _next_nonoverlapping(df_to_find_nearest_in.End, ocdf.Start)
#         elif how == "previous":
#             l_idx, r_idx, dist = _previous_nonoverlapping(df_to_find_nearest_in.Start, ocdf.End)
#         else:
#             previous_l_idx, previous_r_idx, previous_dist = _previous_nonoverlapping(df_to_find_nearest_in.Start, ocdf.End)

#             next_l_idx, next_r_idx, next_dist = _next_nonoverlapping(df_to_find_nearest_in.End, ocdf.Start)

#             l_idx, r_idx, dist = nearest_nonoverlapping(previous_l_idx, previous_r_idx,
#                                                 previous_dist,
#                                                 next_l_idx, next_r_idx, next_dist)

#         ocdf = ocdf.reindex(r_idx, fill_value=-1) # instead of np.nan, so ints are not promoted to float

#         ocdf.index = df_to_find_nearest_in.index
#         ocdf.insert(ocdf.shape[1], "Distance", pd.Series(dist, index=ocdf.index).fillna(-1).astype(int))

#         r_idx = pd.Series(r_idx, index=ocdf.index)
#         df_to_find_nearest_in = df_to_find_nearest_in.drop(r_idx.loc[r_idx == -1].index)

#         df = df_to_find_nearest_in.join(ocdf, rsuffix=suffix)

#     if overlap and "df" in locals() and not df.empty and not nearest_df.empty:
#         df = pd.concat([nearest_df, df])
#     elif overlap and not nearest_df.empty:
#         df = nearest_df

#     # df = df.drop("Chromosome" + suffix, axis=1)
#     return df




# def _next_nonoverlapping(left_ends, right_starts):

#     # print("le", left_ends)
#     # print("rs", right_starts)

#     left_ends = left_ends.sort_values()
#     right_starts = right_starts.sort_values()
#     lidx = left_ends.index.values
#     ridx = right_starts.index.values
#     l_idx, r_idx, dist = nearest_next_nonoverlapping(left_ends.values - 1, lidx, right_starts.values, ridx)
#     r_idx = pd.Series(r_idx, index=left_ends.index).sort_index().values
#     dist = pd.Series(dist, index=left_ends.index).sort_index().values

#     return l_idx, r_idx, dist


# def _previous_nonoverlapping(left_starts, right_ends):

#     left_starts = left_starts.sort_values()
#     right_ends = right_ends.sort_values()
#     l_idx, r_idx, dist = nearest_previous_nonoverlapping(left_starts.values, left_starts.index.values, right_ends.values - 1, right_ends.index.values)
#     # print("ridx before", r_idx)
#     r_idx = pd.Series(r_idx, index=left_starts.index).sort_index().values
#     dist = pd.Series(dist, index=left_starts.index).sort_index().values
#     # print("ridx after", r_idx)

#     return l_idx, r_idx, dist

# # from memory_profiler import profile


# def _both_indexes(self, other, how):

#     # print("self", self)
#     # print("other", other)

#     starts = self.Start.values
#     ends = self.End.values
#     indexes = self.index.values

#     it = NCLS(other.Start.values, other.End.values, other.index.values)

#     if not how:
#         _self_indexes, _other_indexes = it.all_overlaps_both(starts, ends, indexes)
#     elif how == "containment":
#         _self_indexes, _other_indexes = it.all_containments_both(starts, ends, indexes)
#     else:
#         _self_indexes, _other_indexes = it.first_overlap_both(starts, ends, indexes)

#     return _self_indexes, _other_indexes


# def _overlapping_for_nearest(self, other, strandedness, suffix):
#     nearest_df = pd.DataFrame(columns="Chromosome Start End Strand".split())

#     sidx, oidx = _both_indexes(self, other, how="first")

#     if not len(oidx) == 0 or not len(sidx) == 0:

#         idxs = sidx
#         missing_overlap = self.index[~self.index.isin(idxs)]
#         # print("missing overlap", missing_overlap)
#         df_to_find_nearest_in = self.reindex(missing_overlap)

#         odf = other.reindex(oidx)
#         odf.index = sidx
#         sdf = self.reindex(sidx)

#         nearest_df = sdf.join(odf, rsuffix=suffix)
#         # print("nearest_df", nearest_df)
#         nearest_df = nearest_df.drop("Chromosome" + suffix, axis=1)
#         # print("nearest_df", nearest_df)
#         nearest_df.insert(nearest_df.shape[1], "Distance", 0)
#     else:
#         df_to_find_nearest_in = self

#     return nearest_df, df_to_find_nearest_in

def _set_union(scdf, ocdf, **kwargs):

    chromosome, strand = parse_grpby_key(kwargs["key"])

    strandedness = kwargs["strandedness"]
    strand = True if strandedness == "same" else False

    if len(scdf) == 0:
        return _cluster(ocdf, chromosome, strand=strand)
    elif len(ocdf) == 0:
        return _cluster(scdf, chromosome, strand=strand)

    _starts = np.concatenate([
        scdf.Start.values,
        ocdf.Start.values])
    _ends = np.concatenate([
        scdf.End.values,
        ocdf.End.values])

    cdf = pd.DataFrame({"Start": _starts, "End": _ends})["Start End".split()]
    cdf = cdf.sort_values("Start")
    starts, ends = find_clusters(cdf.Start.values, cdf.End.values)

    chromosome = scdf.head(1)["Chromosome"].iloc[0]
    if strandedness == "same":
        _strand = scdf.head(1)["Strand"].iloc[0]
    else:
        _strand = False

    cluster_df = _create_df_from_starts_ends(starts, ends, chromosome, _strand)

    return cluster_df


@ray.remote
def _subtraction(scdf, ocdf, kwargs):

    if ocdf.empty or scdf.empty:
        return scdf

    strandedness = kwargs["strandedness"]
    strand = True if strandedness else False

    chromosome = scdf.Chromosome.head(1).iloc[0]
    kwargs["chromosome"] = chromosome

    if "Strand" in ocdf and strand:
        strand = scdf.Strand.head(1).iloc[0]
        kwargs["strand"] = strand

    oc = ray.get(_cluster.remote(ocdf, kwargs))
    # print("oc")
    # print(oc.head())
    # print(oc.dtypes)
    # print(oc.index.values.dtype)
    o = NCLS(oc.Start.values, oc.End.values, oc.index.values)
    # print(o)

    # print("s")
    # print(scdf.head())
    # print(scdf.dtypes)
    # print(scdf.index.values.dtype)
    idx_self, new_starts, new_ends = o.set_difference_helper(
        scdf.Start.values,
        scdf.End.values,
        scdf.index.values)

    missing_idx = pd.Index(scdf.index).difference(idx_self)

    idx_to_drop = new_starts != -1

    new_starts = new_starts[idx_to_drop]
    new_ends = new_ends[idx_to_drop]

    idx_self = idx_self[idx_to_drop]
    new_starts = pd.Series(new_starts, index=idx_self)#.sort_index()
    new_ends = pd.Series(new_ends, index=idx_self)#.sort_index()
    # idx_self = np.sort(idx_self)

    scdf = scdf.reindex(missing_idx.union(idx_self))

    if len(idx_self):
        scdf.loc[scdf.index.isin(idx_self), "Start"] = new_starts
        scdf.loc[scdf.index.isin(idx_self), "End"] = new_ends

    if not scdf.empty:
        return scdf
    else:
        return None


def _coverage(ranges, value_col=None, strand=True):

    try:
        from pyrle.methods import coverage
        from pyrle import PyRles
    except ImportError:
        raise Exception("Using the coverage method requires that pyrle is installed.")

    kwargs = {"value_col": value_col}
    return PyRles(pyrange_apply_single(coverage, ranges, strand, kwargs))



@ray.remote
def _write_both(scdf, ocdf, kwargs):

    if scdf.empty or ocdf.empty:
        return None

    suffixes = kwargs["suffixes"]
    suffix = kwargs["suffix"]
    how = kwargs["how"]
    new_pos = kwargs["new_pos"]

    scdf, ocdf = _both_dfs(scdf, ocdf, how=how)
    nix = pd.Index(range(len(scdf)))
    scdf.index = nix
    ocdf.index = nix

    ocdf = ocdf.drop("Chromosome", axis=1)

    if not new_pos:
        df = scdf.join(ocdf, rsuffix=suffixes[1])

    elif new_pos == "intersection":

        new_starts = pd.Series(
            np.where(scdf.Start.values > ocdf.Start.values, scdf.Start, ocdf.Start),
            index=scdf.index, dtype=np.long)

        new_ends = pd.Series(
            np.where(scdf.End.values < ocdf.End.values, scdf.End, ocdf.End),
            index=scdf.index, dtype=np.long)
        df = scdf.join(ocdf, lsuffix=suffixes[0], rsuffix=suffixes[1])
        df.insert(1, "Start", new_starts)
        df.insert(2, "End", new_ends)
        df.rename(index=str, columns={"Chromosome" + suffixes[0]: "Chromosome", "Strand" + suffixes[0]: "Strand"}, inplace=True)

    elif new_pos == "union":

        new_starts = pd.Series(
            np.where(scdf.Start.values < ocdf.Start.values, scdf.Start, ocdf.Start),
            index=scdf.index, dtype=np.long)

        new_ends = pd.Series(
            np.where(scdf.End.values > ocdf.End.values, scdf.End, ocdf.End),
            index=scdf.index, dtype=np.long)
        df = scdf.join(ocdf, lsuffix=suffixes[0], rsuffix=suffixes[1])
        df.insert(1, "Start", new_starts)
        df.insert(2, "End", new_ends)
        df.rename(index=str, columns={"Chromosome" + suffixes[0]: "Chromosome", "Strand" + suffixes[0]: "Strand"}, inplace=True)

    return df

def _lengths(df):

    lengths = df.End - df.Start

    return lengths

@ray.remote
def _jaccard(self, other, kwargs):

    strandedness = kwargs["strandedness"]

    if strandedness:
        strand = True
    else:
        strand = False

    s = _lengths(self).sum()
    o = _lengths(other).sum()

    res = ray.get(_intersection.remote(self, other, kwargs))
    if isinstance(res, pd.DataFrame):
        if not res.empty:
            il = _lengths(res).sum()
        else:
            il = 0
    else:
        il = 0

    return [ s, o, il ]


@ray.remote
def _tss(df, kwargs):

    slack = kwargs["slack"]

    tss_pos = df.loc[df.Strand == "+"]

    tss_neg = df.loc[df.Strand == "-"].copy()

    # pd.options.mode.chained_assignment = None
    tss_neg.loc[:, "Start"] = tss_neg.End

    # pd.options.mode.chained_assignment = "warn"
    tss = pd.concat([tss_pos, tss_neg], sort=False)
    tss["End"] = tss.Start
    tss.End = tss.End + 1 + slack
    tss.Start = tss.Start - slack
    tss.loc[tss.Start < 0, "Start"] = 0

    return tss.reindex(df.index)


@ray.remote
def _tes(df, kwargs):

    slack = kwargs["slack"]

    tes_pos = df.loc[df.Strand == "+"]

    tes_neg = df.loc[df.Strand == "-"].copy()

    # pd.options.mode.chained_assignment = None
    tes_neg.loc[:, "Start"] = tes_neg.End

    # pd.options.mode.chained_assignment = "warn"
    tes = pd.concat([tes_pos, tes_neg], sort=False)
    tes["Start"] = tes.End
    tes.End = tes.End + 1 + slack
    tes.Start = tes.Start - slack
    tes.loc[tes.Start < 0, "Start"] = 0

    return tes.reindex(df.index)


@ray.remote
def _slack(df, kwargs):

    slack = kwargs["slack"]
    # df = self.df.copy()
    df.Start = df.Start - slack
    df.loc[df.Start < 0, "Start"] = 0
    df.End = df.End + slack

    return df

@ray.remote
def _sort(df, kwargs):

    return sort_one_by_one(df, "Start", "End")


if __name__ == "__main__":

    chip_f = "/mnt/scratch/endrebak/pyranges_benchmark/data/download/chip_1000000.bed.gz"
    background_f = "/mnt/scratch/endrebak/pyranges_benchmark/data/download/input_1000000.bed.gz"


    chip = pd.read_table(chip_f, sep="\t", usecols=[0, 1, 2, 5], header=None,
                            names="Chromosome Start End Strand".split(),
                            dtype={"Chromosome": "category", "Strand": "category"} )

    cgr = pr.PyRanges(chip, copy_df=False)

    background = pd.read_table(background_f, sep="\t", usecols=[0, 1, 2, 5],
                                header=None, names="Chromosome Start End Strand".split(),
                                dtype={"Chromosome": "category", "Strand": "category"})

    bgr = pr.PyRanges(background, copy_df=False)



# import pandas as pd
# import numpy as np

# import itertools

# # import multiprocessing
# # from multiprocessing import Process
# # from joblib import Parallel, delayed

# #nearest_next,
# # nearest_previous)(nearest, nearest_nonoverlapping,

# from collections import OrderedDict

# from sorted_nearest import (nearest_previous_nonoverlapping,
#                             nearest_next_nonoverlapping,
#                             nearest_nonoverlapping, find_clusters)
# try:
#     dummy = profile
# except:
#     profile = lambda x: x


# # from clustertree import find_clusters

# import pyranges as pr

# from time import time
# import datetime

# from collections import defaultdict

# def pick_out_indexes_possibly_nonunique(df, indexes, invert=False):

#     if isinstance(indexes, list) and indexes:
#         concat = np.concatenate(indexes)
#         indexes = np.unique(concat)

#     if not invert:
#         return df.loc[df.index.isin(indexes)]
#     else:
#         return df.loc[~df.index.isin(indexes)]


# def _overlap(self, other, strandedness, invert, how=None):

#     assert strandedness in ["same", "opposite", False, None]

#     df = self.df
#     other_strand = {"+": "-", "-": "+"}
#     same_strand = {"+": "+", "-": "-"}

#     if strandedness == "same":
#         strand_dict = same_strand
#     elif strandedness == "opposite":
#         strand_dict = other_strand

#     dfs = OrderedDict()

#     if self.stranded and other.stranded and strandedness:

#         indexes_of_overlapping = []
#         for (chromosome, strand), cdf in df.groupby("Chromosome Strand".split()):

#             starts = cdf.Start.values
#             ends = cdf.End.values
#             indexes = cdf.index.values

#             strand = strand_dict[strand]
#             it = other.__ncls__[chromosome, strand]
#             if how == "containment":
#                 indexes = it.has_containment(starts, ends, indexes)
#             else:
#                 indexes = it.has_overlaps(starts, ends, indexes)


#             dfs[chromosome, strand] = pick_out_indexes_possibly_nonunique(cdf, indexes, invert)

#     else:

#         for chromosome, cdf in df.groupby("Chromosome"):

#             starts = cdf.Start.values
#             ends = cdf.End.values
#             indexes = cdf.index.values

#             it = other.__ncls__[chromosome, "+"]
#             it2 = other.__ncls__[chromosome, "-"]

#             if how == "containment":
#                 idx1 = it.has_containments(starts, ends, indexes)
#                 idx2 = it2.has_containments(starts, ends, indexes)
#             else:
#                 idx1 = it.has_overlaps(starts, ends, indexes)
#                 idx2 = it2.has_overlaps(starts, ends, indexes)

#             dfs[chromosome] = pick_out_indexes_possibly_nonunique(cdf, [idx1, idx2], invert)

#     return pd.concat(dfs.values())


# def both_indexes(self, other, strandedness, how=None):

#     assert strandedness in ["same", "opposite", False, None]
#     assert how in "containment first".split() + [False, None]

#     if strandedness:
#         assert self.stranded and other.stranded, \
#             "Can only do stranded searches when both PyRanges contain strand info"

#     df = self.df

#     other_strand = {"+": "-", "-": "+"}
#     same_strand = {"+": "+", "-": "-"}

#     if strandedness == "same":
#         strand_dict = same_strand
#     elif strandedness == "opposite":
#         strand_dict = other_strand

#     self_d = OrderedDict()
#     other_d = OrderedDict()

#     if how == "first" and not strandedness and other.stranded:
#         other_dfs = OrderedDict([(k, v) for k, v in other.df.groupby("Chromosome")])

#     if strandedness: # know from assertion above that both have strand info

#         for (chromosome, strand), cdf in df.groupby("Chromosome Strand".split()):

#             starts = cdf.Start.values
#             ends = cdf.End.values
#             indexes = cdf.index.values

#             strand = strand_dict[strand]
#             it = other.__ncls__[chromosome, strand]

#             if not how:
#                 _self_indexes, _other_indexes = it.all_overlaps_both(starts, ends, indexes)
#             elif how == "containment":
#                 _self_indexes, _other_indexes = it.all_containments_both(starts, ends, indexes)
#             else:
#                 _self_indexes, _other_indexes = it.first_overlap_both(starts, ends, indexes)


#             self_d[chromosome, strand] = _self_indexes
#             other_d[chromosome, strand] = _other_indexes

#     # if not strandedness, and other df has strand, need to search both ncls
#     elif other.stranded:

#         # print("elif other stranded " * 10, )
#         # print("self\n", self)
#         # print("other\n", other)
#         for chromosome, cdf in df.groupby("Chromosome"):

#             it = other.__ncls__[chromosome, "+"]
#             it2 = other.__ncls__[chromosome, "-"]

#             starts = cdf.Start.values
#             ends = cdf.End.values
#             indexes = cdf.index.values

#             if not how:
#                 # print("all overlaps " * 10)
#                 _self_indexes, _other_indexes = it.all_overlaps_both(starts, ends, indexes)
#                 _self_indexes2, _other_indexes2 = it2.all_overlaps_both(starts, ends, indexes)
#             elif how == "containment":
#                 # print("all containments " * 10)
#                 _self_indexes, _other_indexes = it.all_containments_both(starts, ends, indexes)
#                 _self_indexes2, _other_indexes2 = it2.all_containments_both(starts, ends, indexes)
#             else:
#                 _self_indexes, _other_indexes = it.first_overlap_both(starts, ends, indexes)
#                 _self_indexes2, _other_indexes2 = it2.first_overlap_both(starts, ends, indexes)

#                 if chromosome in other_dfs:
#                     odf = other_dfs[chromosome]
#                 else:
#                     continue

#                 if not len(_other_indexes2):
#                     pass
#                 elif not len(_other_indexes):
#                     other_indexes = _other_indexes2
#                 else:
#                     s = pd.Series(np.concatenate([_self_indexes, _self_indexes2])).duplicated(keep=False)
#                     _dup1, _dup2 = s[:len(_self_indexes)].values, s[len(_self_indexes):].values
#                     # print("dup1", _dup1)
#                     # print("dup2", _dup2)
#                     # print(_other_indexes[_dup1])
#                     # print(_other_indexes2[_dup2])
#                     cond = odf.loc[_other_indexes[_dup1]].Start.values < odf.loc[_other_indexes2[_dup2]].Start.values
#                     # print("os", odf.loc[_other_indexes[_dup1]].Start.values)
#                     # print("os2", odf.loc[_other_indexes2[_dup2]].Start.values)
#                     # print("cond", cond)

#                     _other_indexes_tmp = np.where(cond, _other_indexes[_dup1], _other_indexes2[_dup2])
#                     _self_indexes_tmp = np.where(cond, _self_indexes[_dup1], _self_indexes2[_dup2])
#                     # print(_self_indexes, ~_dup1)
#                     # print("_self_indexes[~_dup1]", _self_indexes[~_dup1])
#                     # print("_self_indexes[~_dup2]", _self_indexes2[~_dup2])
#                     # print("_other_indexes[~_dup1]", _other_indexes[~_dup1])
#                     # print("_other_indexes[~_dup2]", _other_indexes2[~_dup2])
#                     # _self_indexes2 = np.concatenate(, _self_indexes2[~_dup2])
#                     _self_indexes2 = np.concatenate([_self_indexes[~_dup1], _self_indexes2[~_dup2]])
#                     _other_indexes2 = np.concatenate([_other_indexes[~_dup1], _other_indexes2[~_dup2]])

#                     _self_indexes = _self_indexes_tmp
#                     _other_indexes = _other_indexes_tmp
#                     # print("~cond", ~cond)
#                     # print("~_dup1", ~_dup1, _dup1)
#                     # print("~_dup2", ~_dup2, _dup2)
#                     # print("~_dup1", _other_indexes[~_dup1], _self_indexes[~_dup1])
#                     # print("~_dup2", _other_indexes[~_dup2], _self_indexes[~_dup2])
#                     # _other_indexes = np.where(~cond, _other_indexes[~_dup1], _other_indexes2[~_dup2])
#                     # _self_indexes = np.where(~cond, _self_indexes[~_dup1], _self_indexes2[~_dup2])

#             self_d[chromosome] = np.concatenate([_self_indexes, _self_indexes2])
#             other_d[chromosome] = np.concatenate([_other_indexes, _other_indexes2])

#             if not len(_self_indexes) or not len(_self_indexes2):
#                 self_d[chromosome] = np.array(self_d[chromosome], dtype=int)
#                 other_d[chromosome] = np.array(other_d[chromosome], dtype=int)


#     # if not strandedness, and other df has no strand info, need to search one ncls
#     elif not other.stranded:

#         for chromosome, cdf in df.groupby("Chromosome"):

#             it = other.__ncls__[chromosome]

#             starts = cdf.Start.values
#             ends = cdf.End.values
#             indexes = cdf.index.values

#             if not how:
#                 _self_indexes, _other_indexes = it.all_overlaps_both(starts, ends, indexes)
#             elif how == "containment":
#                 _self_indexes, _other_indexes = it.all_containments_both(starts, ends, indexes)
#             else:
#                 _self_indexes, _other_indexes = it.first_overlap_both(starts, ends, indexes)

#             self_d[chromosome] = _self_indexes
#             other_d[chromosome] = _other_indexes

#     return self_d, other_d


# def sort_one_by_one(d, col1, col2):
#     """
#     Equivalent to pd.sort_values(by=[col1, col2]), but faster.
#     """
#     d = d.sort_values(by=[col2])
#     return d.sort_values(by=[col1], kind='mergesort')


# def _cluster(self, strand=False, maxdist=0, minnb=1):

#     dfs = []

#     idx_start, idx_end = 0, 0
#     if strand:

#         for (c, s), cdf in self.df.groupby(["Chromosome", "Strand"]):
#             cdf = cdf.sort_values("Start")
#             starts, ends = find_clusters(cdf.Start.values, cdf.End.values)
#             df = pd.DataFrame({"Chromosome": c, "Start": starts, "End": ends, "Strand": s})
#             dfs.append(df)

#         df = pd.concat(dfs, ignore_index=True)["Chromosome Start End Strand".split()]

#     else:

#         for c, cdf in self.df.groupby(["Chromosome"]):
#             cdf = cdf.sort_values("Start")
#             starts, ends = find_clusters(cdf.Start.values, cdf.End.values)
#             df = pd.DataFrame({"Chromosome": c, "Start": starts, "End": ends})
#             dfs.append(df)

#         df = pd.concat(dfs, ignore_index=True)["Chromosome Start End".split()]


#     return df



# def _tile(self, tile_size=50):

#     "No no no this is slow! Write in C."

#     df = self.df.copy()

#     df.Start = df.Start - (df.Start % tile_size)
#     df.End = df.End - (df.End % tile_size) + tile_size - 1

#     rows = []
#     for _, row in df.iterrows():
#         d = row.to_dict()
#         for tile in range(row.Start, row.End, tile_size):
#             d2 = d.copy()
#             d2["Start"], d2["End"] = tile, tile + tile_size - 1
#             rows.append(d2)

#     df = pd.DataFrame.from_dict(rows)[df.columns]

#     return df


# def invert(self):

#     pass


# # def _multithreaded_intersection(key, scdf, ocdf, strandedness=None, how=None):


# #     # is there anything to gain?

# #     # create ncls

# #     # both indexes

# #     # intersect op

# #     # c, s = list(ocdf.head(0)["Chromosome Strand".split()].values)

# #     starts = np.copy(scdf.Start.values)
# #     ends = np.copy(scdf.End.values)
# #     indexes = np.copy(scdf.index.values)

# #     it = pr.PyRanges(ocdf).__ncls__[key]

# #     _self_indexes, _other_indexes = it.all_overlaps_both(starts, ends, indexes)

# #     scdf = scdf.loc[_self_indexes]
# #     ocdf = ocdf.loc[_other_indexes, ["Start", "End"]]

# #     new_starts = pd.Series(
# #         np.where(scdf.Start.values > ocdf.Start.values, scdf.Start, ocdf.Start),
# #         index=scdf.index, dtype=np.long)

# #     new_ends = pd.Series(
# #         np.where(scdf.End.values < ocdf.End.values, scdf.End, ocdf.End),
# #         index=scdf.index, dtype=np.long)

# #     pd.options.mode.chained_assignment = None  # default='warn'
# #     scdf.loc[:, "Start"] = new_starts
# #     scdf.loc[:, "End"] = new_ends
# #     pd.options.mode.chained_assignment = 'warn'

# #     return scdf




# def _intersection(self, other, strandedness=None, how=None):

#     sidx, oidx = both_indexes(self, other, strandedness, how)

#     other_strand = {"+": "-", "-": "+"}

#     if len(list(sidx.keys())[0]) == 2: # chromosome and strand
#         grpby_key = "Chromosome Strand".split()
#     else:
#         grpby_key = "Chromosome"

#     other_dfs = {k: d for k, d in other.df.groupby(grpby_key)}

#     dfs = []

#     for key, scdf in self.df.groupby(grpby_key):

#         if len(key) == 2 and strandedness == "opposite":
#             key = key[0], other_strand[key[1]]

#         if not key in other_dfs:
#             continue

#         ocdf = other_dfs[key]

#         scidx = sidx[key]
#         ocidx = oidx[key]

#         scdf = scdf.loc[scidx]
#         ocdf = ocdf.loc[ocidx, ["Start", "End"]]

#         new_starts = pd.Series(
#             np.where(scdf.Start.values > ocdf.Start.values, scdf.Start, ocdf.Start),
#             index=scdf.index, dtype=np.long)

#         new_ends = pd.Series(
#             np.where(scdf.End.values < ocdf.End.values, scdf.End, ocdf.End),
#             index=scdf.index, dtype=np.long)

#         pd.options.mode.chained_assignment = None  # default='warn'
#         scdf.loc[:, "Start"] = new_starts
#         scdf.loc[:, "End"] = new_ends
#         pd.options.mode.chained_assignment = 'warn'

#         dfs.append(scdf)

#     if dfs:
#         df = pd.concat(dfs)
#     else:
#         df = pd.DataFrame(columns="Chromosome Start End Strand".split())

#     return df



# def _coverage(ranges, value_col=None, stranded=False):

#     try:
#         from pyranges import PyRles
#     except ImportError:
#         raise Exception("Using the coverage method requires that pyrle is installed.")

#     return PyRles(ranges, value_col=value_col, stranded=stranded)


# def _keep_both(idx_df, other, suffixes, new_pos):

#     self_suffix, other_suffix = suffixes

#     if new_pos:
#         assert self_suffix and other_suffix, "If new_pos is union or intersection, you must give two suffixes of length > 0."
#     else:
#         assert other_suffix, "Must give a suffix for the right dataframe."

#     if not new_pos:
#         suffixes = ["", other_suffix]
#         to_drop = ["IndexOther", "Chromosome" + other_suffix]
#         self_suffix = ""
#         to_rename = {}
#     else:
#         suffixes = [self_suffix, other_suffix]
#         to_drop = ["IndexOther", "Chromosome" + other_suffix]
#         to_rename = {"Chromosome" + self_suffix: "Chromosome"}

#     idx_df = idx_df.merge(other.df, left_on="IndexOther", right_index=True, suffixes=suffixes)
#     idx_df = idx_df.drop(to_drop, axis=1)
#     idx_df = idx_df.rename(to_rename, axis="columns")

#     if new_pos == "intersect":
#         new_starts = np.where(idx_df["Start" + self_suffix] > idx_df["Start" + other_suffix], idx_df["Start" + self_suffix], idx_df["Start" + other_suffix])
#         new_ends = np.where(idx_df["End" + self_suffix] < idx_df["End" + other_suffix], idx_df["End" + self_suffix], idx_df["End" + other_suffix])
#         idx_df.insert(1, "Start", new_starts)
#         idx_df.insert(2, "End", new_ends)

#     if new_pos == "union":
#         new_starts = np.where(idx_df["Start" + self_suffix] < idx_df["Start" + other_suffix], idx_df["Start" + self_suffix], idx_df["Start" + other_suffix])
#         new_ends = np.where(idx_df["End" + self_suffix] > idx_df["End" + other_suffix], idx_df["End" + self_suffix], idx_df["End" + other_suffix])
#         idx_df.insert(1, "Start", new_starts)
#         idx_df.insert(2, "End", new_ends)

#     return idx_df


# def _overlap_write_both(self, other, strandedness=False, new_pos=None, suffixes=["_a", "_b"], suffix="_b", how=None):

#     assert new_pos in ["intersection", "union", False, None]

#     sidx, oidx = both_indexes(self, other, strandedness, how)

#     other_strand = {"+": "-", "-": "+"}

#     if len(list(sidx.keys())[0]) == 2: # chromosome and strand
#         grpby_key = "Chromosome Strand".split()
#     else:
#         grpby_key = "Chromosome"

#     other_dfs = {k: d for k, d in other.df.groupby(grpby_key)}

#     dfs = []

#     idx_counter, current_end = 0, 0
#     for key, scdf in self.df.groupby(grpby_key):

#         if len(key) == 2 and strandedness == "opposite":
#             key = key[0], other_strand[key[1]]

#         if not key in other_dfs:
#             continue

#         ocdf = other_dfs[key].drop("Chromosome", 1)

#         scidx = sidx[key]
#         ocidx = oidx[key]

#         scdf = scdf.loc[scidx]
#         ocdf = ocdf.loc[ocidx]

#         pd.options.mode.chained_assignment = None  # default='warn'
#         pd.options.mode.chained_assignment = 'warn'

#         current_end += len(scdf)
#         scdf.index = range(idx_counter, current_end)
#         ocdf.index = range(idx_counter, current_end)
#         idx_counter += len(scdf)

#         if not new_pos:
#             _df = scdf.join(ocdf, rsuffix=suffix)

#         elif new_pos == "intersection":

#             new_starts = pd.Series(
#                 np.where(scdf.Start.values > ocdf.Start.values, scdf.Start, ocdf.Start),
#                 index=scdf.index, dtype=np.long)

#             new_ends = pd.Series(
#                 np.where(scdf.End.values < ocdf.End.values, scdf.End, ocdf.End),
#                 index=scdf.index, dtype=np.long)
#             _df = scdf.join(ocdf, lsuffix=suffixes[0], rsuffix=suffixes[1])
#             _df.insert(1, "Start", new_starts)
#             _df.insert(2, "End", new_ends)
#             _df.rename(index=str, columns={"Chromosome" + suffixes[0]: "Chromosome", "Strand" + suffixes[0]: "Strand"}, inplace=True)

#         elif new_pos == "union":

#             new_starts = pd.Series(
#                 np.where(scdf.Start.values < ocdf.Start.values, scdf.Start, ocdf.Start),
#                 index=scdf.index, dtype=np.long)

#             new_ends = pd.Series(
#                 np.where(scdf.End.values > ocdf.End.values, scdf.End, ocdf.End),
#                 index=scdf.index, dtype=np.long)
#             _df = scdf.join(ocdf, lsuffix=suffixes[0], rsuffix=suffixes[1])
#             _df.insert(1, "Start", new_starts)
#             _df.insert(2, "End", new_ends)
#             _df.rename(index=str, columns={"Chromosome" + suffixes[0]: "Chromosome", "Strand" + suffixes[0]: "Strand"}, inplace=True)

#         dfs.append(_df)


#     if dfs:
#         df = pd.concat(dfs)
#     else:
#         return pd.DataFrame(columns="Chromosome Start End Strand".split())

#     return df


# def _set_intersection(self, other, strandedness=None, how=None):

#     strand = True if strandedness else False
#     s = self.cluster(strand=strand)
#     o = other.cluster(strand=strand)

#     return _intersection(s, o, strandedness, how)


# def _set_union(self, other, strand):

#     if strand:
#         assert self.stranded and other.stranded, \
#             "Can only do stranded set union when both PyRanges contain strand info."

#     if len(self) == 0:
#         return _cluster(other, strand=strand)
#     elif len(other) == 0:
#         return _cluster(self, strand=strand)

#     dfs = []

#     if strand and len(list(self.__ncls__.keys())[0]) == 2: # chromosome and strand
#         grpby_key = "Chromosome Strand".split()
#         columns = "Chromosome Start End Strand".split()
#     else:
#         grpby_key = "Chromosome"
#         columns = "Chromosome Start End".split()

#     self_dfs =  {k: d for k, d in self.df.groupby(grpby_key)}
#     other_dfs = {k: d for k, d in other.df.groupby(grpby_key)}

#     idx_start, idx_end = 0, 0
#     for key in set(self_dfs).union(other_dfs):

#         if key in other_dfs and key in self_dfs:
#             _starts = np.concatenate([
#                 self_dfs[key].Start.values,
#                 other_dfs[key].Start.values])
#             _ends = np.concatenate([
#                 self_dfs[key].End.values,
#                 other_dfs[key].End.values])
#         elif key in self_dfs and not key in other_dfs:
#             _starts = self_dfs[key].Start.values
#             _ends = self_dfs[key].End.values
#         elif key in other_dfs and not key in self_dfs:
#             _starts = other_dfs[key].Start.values
#             _ends = other_dfs[key].End.values


#         cdf = pd.DataFrame({"Start": _starts, "End": _ends})["Start End".split()]
#         # print("cdf" * 10 + "\n", cdf)
#         # cdf = sort_one_by_one(cdf, "Start", "End")
#         cdf = cdf.sort_values("Start")
#         # print("cdf" * 10 + "\n", cdf)
#         # print(clusters)
#         starts, ends = find_clusters(cdf.Start.values, cdf.End.values)
#         # print(starts, ends)
#         idx_end += len(starts)

#         if strand:
#             _df = pd.DataFrame({"Chromosome": key[0], "Start": starts, "End": ends, "Strand": key[1]})
#         else:
#             _df = pd.DataFrame({"Chromosome": key, "Start": starts, "End": ends})

#         _df.index = range(idx_start, idx_end)
#         dfs.append(_df)
#         idx_start = idx_end

#     df = pd.concat(dfs)[columns]

#     return df


# def _subtraction(self, other, strandedness):

#     #print(self.df.to_csv(sep=" "))
#     strand = False
#     if strandedness:
#         assert "Strand" in self.df and "Strand" in other.df, \
#             "Can only do stranded set difference when both PyRanges contain strand info."
#         strand = True

#     if len(self) == 0:
#         return pd.DataFrame(columns="Chromosome Start End Strand".split())
#     if len(other) == 0:
#         return self.df

#     other = other.cluster(strand=strand)

#     other_strand = {"+": "-", "-": "+"}
#     same_strand = {"+": "+", "-": "-"}

#     if strandedness == "same":
#         strand_dict = same_strand
#     elif strandedness == "opposite":
#         strand_dict = other_strand

#     if strandedness and self.stranded: # chromosome and strand
#         grpby_key = "Chromosome Strand".split()
#     else:
#         grpby_key = "Chromosome"

#     # print("self.df", self.df.to_csv(sep=" "))


#     self_dfs = {k: d for k, d in self.df.groupby(grpby_key)}
#     other_dfs = {k: d for k, d in other.df.groupby(grpby_key)}

#     dfs = []

#     # need to store indexes in dict since if it is opposite stranded
#     # cannot do removal in same iteration
#     _idxes = dict()

#     for key, scdf in self_dfs.items():

#         # the downstream union of self/missing indexes sorts them
#         # therefore, we need them to be in correct order here
#         scdf.index = pd.Index(range(len(scdf)))
#         print("-------")
#         print("scdf", scdf.to_csv(sep=" "))


#         if len(key) == 2:
#             old_key = key
#             key = key[0], strand_dict[key[1]]


#         if not key in other_dfs:
#             dfs.append(scdf)
#             continue

#         # scdf = self_dfs[key]
#         ogr = pr.PyRanges(other_dfs[key])
#         # ogr.index = pd.Index(range(len(ogr)))
#         #print("scdf\n", scdf)
#         #print("odf\n", ogr)

#         # check which self reads have partial overlap
#         if not ogr.stranded and not strandedness:
#             #print("1" * 100)
#             idx_self, new_starts, new_ends = ogr.__ncls__[key].set_difference_helper(
#                scdf.Start.values,
#                 scdf.End.values,
#                 scdf.index.values)

#         elif self.stranded and ogr.stranded and strandedness:

#             idx_self, new_starts, new_ends = ogr.__ncls__[key].set_difference_helper(
#                 scdf.Start.values,
#                 scdf.End.values,
#                 scdf.index.values)

#         elif ogr.stranded and not strandedness:

#             #print("3" * 100)
#             idx_self1, new_starts1, new_ends1 = ogr.__ncls__[key, "+"].set_difference_helper(
#                 scdf.Start.values,
#                 scdf.End.values,
#                 scdf.index.values)
#             idx_self2, new_starts2, new_ends2 = ogr.__ncls__[key, "-"].set_difference_helper(
#                 scdf.Start.values,
#                 scdf.End.values,
#                 scdf.index.values)

#             idx_self = np.concatenate([idx_self1, idx_self2])

#             if not idx_self.dtype == np.int64:
#                 idx_self = np.array(idx_self, dtype=np.int64)

#             new_starts = np.concatenate([new_starts1, new_starts2])
#             new_ends = np.concatenate([new_ends1, new_ends2])

#         #print("scdf.index", scdf.index)
#         missing_idx = pd.Index(scdf.index).difference(idx_self)
#         #print("missing_idx " * 10, missing_idx)
#         idx_to_drop = new_starts != -1
#         #print(idx_to_drop)
#         new_starts = new_starts[idx_to_drop]
#         new_ends = new_ends[idx_to_drop]

#         idx_self = idx_self[idx_to_drop]
#         new_starts = pd.Series(new_starts, index=idx_self).sort_index()
#         new_ends = pd.Series(new_ends, index=idx_self).sort_index()
#         # idx_self = np.sort(idx_self)

#         print("scdf", scdf.to_csv(sep=" "))
#         print("scdf.index", scdf.index)
#         print("idx_self", idx_self)
#         print("missing_idx", missing_idx)
#         scdf = scdf.reindex(missing_idx.union(idx_self))
#         print("scdf", scdf.to_csv(sep=" "))
#         print("scdf.index", scdf.index)
#         print("idx_self", idx_self)

#         if len(idx_self):
#             # why does boolean indexing work, but not regular?

#             # pd.options.mode.chained_assignment = None  # default='warn'
#             #print("idx_self", idx_self)
#             scdf.loc[scdf.index.isin(idx_self), "Start"] = new_starts
#             scdf.loc[scdf.index.isin(idx_self), "End"] = new_ends

#         # find those in scdf that were not in idx_self

#         #print("scdf", scdf)

#         dfs.append(scdf)

#     if dfs:
#         df = pd.concat(dfs)
#     else:
#         return pd.DataFrame(columns="Chromosome Start End Strand".split())

#     #print("dfs", dfs)
#     #print("df", df)
#     return df


# # def _next_nonoverlapping(left_ends, right_starts):

# #     left_ends = left_ends.sort_values()
# #     right_starts = right_starts.sort_values()
# #     l_idx, r_idx, dist = nearest_next_nonoverlapping(left_ends.values - 1, left_ends.index.values, right_starts.values, right_starts.index.values)

# #     return l_idx, r_idx, dist


# # def _previous_nonoverlapping(left_starts, right_ends):

# #     left_starts = left_starts.sort_values()
# #     right_ends = right_ends.sort_values()
# #     l_idx, r_idx, dist = nearest_previous_nonoverlapping(left_starts.values, left_starts.index.values, right_ends.values - 1, right_ends.index.values)

# #     return l_idx, r_idx, dist

# def _next_nonoverlapping(left_ends, right_starts, right_indexes):

#     left_ends = left_ends.sort_values()
#     right_starts = right_starts.sort_values()
#     r_idx, dist = nearest_next_nonoverlapping(left_ends.values - 1, right_starts.values, right_indexes)
#     r_idx = pd.Series(r_idx, index=left_ends.index).sort_index().values
#     dist = pd.Series(dist, index=left_ends.index).sort_index().values

#     return r_idx, dist


# def _previous_nonoverlapping(left_starts, right_ends, right_indexes):

#     left_starts = left_starts.sort_values()
#     right_ends = right_ends.sort_values()
#     r_idx, dist = nearest_previous_nonoverlapping(left_starts.values, right_ends.values - 1, right_ends.index.values)
#     # print("ridx before", r_idx)
#     r_idx = pd.Series(r_idx, index=left_starts.index).sort_index().values
#     dist = pd.Series(dist, index=left_starts.index).sort_index().values
#     # print("ridx after", r_idx)

#     return r_idx, dist

# # from memory_profiler import profile


# def _overlapping_for_nearest(self, other, strandedness, suffix):
#     nearest_df = pd.DataFrame(columns="Chromosome Start End Strand".split())

#     sidx, oidx = both_indexes(self, other, strandedness, how="first")

#     if list(oidx.values()):
#         idxs = np.concatenate(list(sidx.values()))
#         missing_overlap = self.df.index[~self.df.index.isin(idxs)]
#         # print("missing overlap", missing_overlap)
#         df_to_find_nearest_in = self.df.reindex(missing_overlap)

#         odf = other.df.reindex(np.concatenate(list(oidx.values())))
#         odf.index = np.concatenate(list(sidx.values()))
#         sdf = self.df.reindex(np.concatenate(list(sidx.values())))

#         nearest_df = sdf.join(odf, rsuffix=suffix).drop("Chromosome" + suffix, axis=1)
#         nearest_df.insert(nearest_df.shape[1], "Distance", 0)
#     else:
#         df_to_find_nearest_in = self.df

#     return nearest_df, df_to_find_nearest_in


# def _nearest(self, other, strandedness, suffix="_b", how=None, overlap=True):

#     if overlap:
#         nearest_df, df_to_find_nearest_in = _overlapping_for_nearest(self, other, strandedness, suffix)
#     else:
#         df_to_find_nearest_in = self.df

#     other_strand = {"+": "-", "-": "+"}

#     if self.stranded and strandedness: # chromosome and strand
#         grpby_key = "Chromosome Strand".split()
#     else:
#         grpby_key = "Chromosome"

#     other_dfs = {k: d for k, d in other.df.groupby(grpby_key)}

#     dfs = []

#     for key, scdf in df_to_find_nearest_in.groupby(grpby_key):

#         if len(key) == 2 and strandedness == "opposite":
#             other_key = key[0], other_strand[key[1]]
#         else:
#             other_key = key

#         if not other_key in other_dfs:
#             continue

#         ocdf = other_dfs[other_key]

#         scdf.index = pd.Index(range(len(scdf)))

#         if how == "next":
#             r_idx, dist = _next_nonoverlapping(scdf.End, ocdf.Start, ocdf.index.values)
#         elif how == "previous":
#             r_idx, dist = _previous_nonoverlapping(scdf.Start, ocdf.End, ocdf.index.values)
#         else:
#             previous_r_idx, previous_dist = _previous_nonoverlapping(scdf.Start, ocdf.End, ocdf.index.values)

#             next_r_idx, next_dist = _next_nonoverlapping(scdf.End, ocdf.Start, ocdf.index.values)

#             r_idx, dist = nearest_nonoverlapping(previous_r_idx,
#                                                  previous_dist,
#                                                  next_r_idx, next_dist)

#         ocdf = ocdf.reindex(r_idx, fill_value=-1) # instead of np.nan, so ints are not promoted to float

#         ocdf.index = scdf.index
#         ocdf.insert(ocdf.shape[1], "Distance", pd.Series(dist, index=ocdf.index).fillna(-1).astype(int))
#         ocdf.drop("Chromosome", axis=1, inplace=True)

#         r_idx = pd.Series(r_idx, index=ocdf.index)
#         scdf = scdf.drop(r_idx.loc[r_idx == -1].index)

#         result = scdf.join(ocdf, rsuffix=suffix)

#         dfs.append(result)

#     if dfs:
#         df = pd.concat(dfs)
#     else:
#         df = pd.DataFrame(columns="Chromosome Start End Strand".split())


#     if overlap and not df.empty and not nearest_df.empty:
#         df = pd.concat([nearest_df, df])
#     elif overlap and not nearest_df.empty:
#         df = nearest_df

#     return df


# def _slack(self, slack):

#     df = self.df.copy()
#     df.Start = df.Start - slack
#     df.loc[df.Start < 0, "Start"] = 0
#     df.End = df.End + slack

#     return df


# def _tss(self, slack=0):

#     df = self.df

#     tss_pos = df.loc[df.Strand == "+"]

#     tss_neg = df.loc[df.Strand == "-"].copy()

#     # pd.options.mode.chained_assignment = None
#     tss_neg.loc[:, "Start"] = tss_neg.End

#     # pd.options.mode.chained_assignment = "warn"
#     tss = pd.concat([tss_pos, tss_neg], sort=False)
#     tss["End"] = tss.Start
#     tss.End = tss.End + 1 + slack
#     tss.Start = tss.Start - slack
#     tss.loc[tss.Start < 0, "Start"] = 0

#     return tss.reindex(df.index)


# def _tes(self, slack=0):

#     df = self.df

#     tes_pos = df.loc[df.Strand == "+"]

#     tes_neg = df.loc[df.Strand == "-"].copy()

#     # pd.options.mode.chained_assignment = None
#     tes_neg.loc[:, "Start"] = tes_neg.End

#     # pd.options.mode.chained_assignment = "warn"
#     tes = pd.concat([tes_pos, tes_neg], sort=False)
#     tes["Start"] = tes.End
#     tes.End = tes.End + 1 + slack
#     tes.Start = tes.Start - slack
#     tes.loc[tes.Start < 0, "Start"] = 0

#     return tes.reindex(df.index)


# def _lengths(self):

#     df = self.df
#     lengths = df.End - df.Start

#     return lengths


# def _jaccard(self, other, strandedness):

#     if strandedness:
#         strand = True
#     else:
#         strand = False

#     il = self.set_intersection(other, strandedness).lengths().sum()
#     # ul = self.set_union(other, strandedness).lengths().sum()
#     s = self.cluster(True).lengths().sum()
#     o = other.cluster(True).lengths().sum()

#     # print("o", o)
#     # print("s", s)
#     # print("il", il)
#     # print("ul", ul)



#     # if il == ul:
#     #     return 1
#     if s + o == il:
#         return 1

#     return il / (s + o - il)
